#!/usr/bin/env python3

import matplotlib
matplotlib.use('Agg')

import os
import sys
import re

import numpy as np
from vmaf.config import DisplayConfig

from vmaf.core.result_store import FileSystemResultStore
from vmaf.tools.misc import import_python_file, get_cmd_option, cmd_option_exists
from vmaf.core.quality_runner import QualityRunner, VmafQualityRunner, BootstrapVmafQualityRunner
from vmaf.core.matlab_quality_runner import STMADQualityRunner, SpEEDMatlabQualityRunner, StrredQualityRunner, StrredOptQualityRunner  # need these to have the runners recognized
from vmaf.routine import run_test_on_dataset, print_matplotlib_warning
from vmaf.tools.stats import ListStats
from vmaf.core.cambi_quality_runner import CambiQualityRunner  # need these to have the runners recognized

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

POOL_METHODS = ['mean', 'harmonic_mean', 'min', 'median', 'perc5', 'perc10', 'perc20']

SUBJECTIVE_MODELS = ['DMOS', 'DMOS_MLE', 'MLE', 'MLE_CO_AP',
                     'MLE_CO_AP2 (default)', 'MOS', 'SR_DMOS',
                     'SR_MOS (i.e. ITU-R BT.500)',
                     'BR_SR_MOS (i.e. ITU-T P.913)',
                     'ZS_SR_DMOS', 'ZS_SR_MOS', '...']

def print_usage():
    quality_runner_types = ['VMAF', 'PSNR', 'SSIM', 'MS_SSIM', '...']
    print("usage: " + os.path.basename(sys.argv[0]) + \
          " quality_type test_dataset_filepath [--vmaf-model VMAF_model_path] " \
          "[--vmaf-phone-model] [--subj-model subjective_model] [--cache-result] " \
          "[--parallelize] [--print-result] [--save-plot plot_dir] [--plot-wh plot_wh] "
          "[--processes processes]\n")
    print("quality_type:\n\t" + "\n\t".join(quality_runner_types) +"\n")
    print("subjective_model:\n\t" + "\n\t".join(SUBJECTIVE_MODELS) + "\n")
    print("plot_wh: plot width and height in inches, example: 5x5 (default)")
    print("processes: must be an integer >=1")


def main():
    if len(sys.argv) < 3:
        print_usage()
        return 2

    try:
        quality_type = sys.argv[1]
        test_dataset_filepath = sys.argv[2]
    except ValueError:
        print_usage()
        return 2

    vmaf_model_path = get_cmd_option(sys.argv, 3, len(sys.argv), '--vmaf-model')
    cache_result = cmd_option_exists(sys.argv, 3, len(sys.argv), '--cache-result')
    parallelize = cmd_option_exists(sys.argv, 3, len(sys.argv), '--parallelize')
    processes = get_cmd_option(sys.argv, 3, len(sys.argv), '--processes')
    print_result = cmd_option_exists(sys.argv, 3, len(sys.argv), '--print-result')
    suppress_plot = cmd_option_exists(sys.argv, 3, len(sys.argv), '--suppress-plot')
    vmaf_phone_model = cmd_option_exists(sys.argv, 3, len(sys.argv), '--vmaf-phone-model')

    pool_method = get_cmd_option(sys.argv, 3, len(sys.argv), '--pool')
    if not (pool_method is None
            or pool_method in POOL_METHODS):
        print('--pool can only have option among {}'.format(', '.join(POOL_METHODS)))
        return 2

    subj_model = get_cmd_option(sys.argv, 3, len(sys.argv), '--subj-model')

    try:
        from sureal.subjective_model import SubjectiveModel
        if subj_model is not None:
            subj_model_class = SubjectiveModel.find_subclass(subj_model)
        else:
            subj_model_class = SubjectiveModel.find_subclass('MLE_CO_AP2')
    except Exception as e:
        print("Error: " + str(e))
        return 1

    save_plot_dir = get_cmd_option(sys.argv, 3, len(sys.argv), '--save-plot')

    plot_wh = get_cmd_option(sys.argv, 3, len(sys.argv), '--plot-wh')
    if plot_wh is not None:
        try:
            mo = re.match(r"([0-9]+)x([0-9]+)", plot_wh)
            assert mo is not None
            w = mo.group(1)
            h = mo.group(2)
            w = int(w)
            h = int(h)
            plot_wh = (w, h)
        except Exception as e:
            print("Error: plot_wh must be in the format of WxH, example: 5x5")
            return 1

    try:
        runner_class = QualityRunner.find_subclass(quality_type)
    except Exception as e:
        print("Error: " + str(e))
        return 1

    if vmaf_model_path is not None and runner_class != VmafQualityRunner and \
                    runner_class != BootstrapVmafQualityRunner:
        print("Input error: only quality_type of VMAF accepts --vmaf-model.")
        print_usage()
        return 2

    if vmaf_phone_model and runner_class != VmafQualityRunner and \
                    runner_class != BootstrapVmafQualityRunner:
        print("Input error: only quality_type of VMAF accepts --vmaf-phone-model.")
        print_usage()
        return 2

    if processes is not None:
        try:
            processes = int(processes)
        except ValueError:
            print("Input error: processes must be an integer")
        assert processes >= 1

    try:
        test_dataset = import_python_file(test_dataset_filepath)
    except Exception as e:
        print("Error: " + str(e))
        return 1

    if cache_result:
        result_store = FileSystemResultStore()
    else:
        result_store = None

    # pooling
    if pool_method == 'harmonic_mean':
        aggregate_method = ListStats.harmonic_mean
    elif pool_method == 'min':
        aggregate_method = np.min
    elif pool_method == 'median':
        aggregate_method = np.median
    elif pool_method == 'perc5':
        aggregate_method = ListStats.perc5
    elif pool_method == 'perc10':
        aggregate_method = ListStats.perc10
    elif pool_method == 'perc20':
        aggregate_method = ListStats.perc20
    else: # None or 'mean'
        aggregate_method = np.mean

    if vmaf_phone_model:
        enable_transform_score = True
    else:
        enable_transform_score = None

    try:
        if suppress_plot:
            raise AssertionError

        from vmaf import plt
        if plot_wh is None:
            plot_wh = (5, 5)
        fig, ax = plt.subplots(figsize=plot_wh, nrows=1, ncols=1)

        assets, results = run_test_on_dataset(test_dataset, runner_class, ax,
                                              result_store, vmaf_model_path,
                                              parallelize=parallelize,
                                              aggregate_method=aggregate_method,
                                              subj_model_class=subj_model_class,
                                              enable_transform_score=enable_transform_score,
                                              processes=processes,
                                              )

        bbox = {'facecolor':'white', 'alpha':0.5, 'pad':20}
        ax.annotate('Testing Set', xy=(0.1, 0.85), xycoords='axes fraction', bbox=bbox)

        # ax.set_xlim([-10, 110])
        # ax.set_ylim([-10, 110])

        plt.tight_layout()

        if save_plot_dir is None:
            DisplayConfig.show()
        else:
            DisplayConfig.show(write_to_dir=save_plot_dir)

    except ImportError:
        print_matplotlib_warning()
        assets, results = run_test_on_dataset(test_dataset, runner_class, None,
                                              result_store, vmaf_model_path,
                                              parallelize=parallelize,
                                              aggregate_method=aggregate_method,
                                              subj_model_class=subj_model_class,
                                              enable_transform_score=enable_transform_score,
                                              processes=processes,
                                              )
    except AssertionError:
        assets, results = run_test_on_dataset(test_dataset, runner_class, None,
                                              result_store, vmaf_model_path,
                                              parallelize=parallelize,
                                              aggregate_method=aggregate_method,
                                              subj_model_class=subj_model_class,
                                              enable_transform_score=enable_transform_score,
                                              processes=processes,
                                              )

    if print_result:
        for result in results:
            print(result)
            print('')

    return 0


if __name__ == '__main__':
    ret = main()
    exit(ret)



__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

"""
Run embedded doctests
"""

import doctest

from vmaf.tools import misc
# from vmaf.tools import stats
from vmaf.core import quality_runner
from vmaf.core import train_test_model

def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(misc))
    # tests.addTests(doctest.DocTestSuite(stats)) # commented out because not numerically exact
    tests.addTests(doctest.DocTestSuite(quality_runner))
    tests.addTests(doctest.DocTestSuite(train_test_model))
    return tests



import numpy as np
cimport numpy as np

from libc.stdlib cimport calloc, free

from vmaf.core.adm_dwt2_tools import ALIGN_CEIL, MAX_ALIGN

cdef extern from "feature/adm_tools.h":
    ctypedef struct adm_dwt_band_t_d:
        double *band_a
        double *band_v
        double *band_h
        double *band_d

cdef extern from "../../../libvmaf/src/feature/adm_tools.c":
    void dwt2_src_indices_filt_s(int **src_ind_y, int **src_ind_x, int w, int h)
    void adm_dwt2_d(const double *src, const adm_dwt_band_t_d *dst, int **ind_y, int **ind_x, int w, int h, int src_stride, int dst_stride)

cdef extern from "../../../libvmaf/src/feature/adm.c":
    char *init_dwt_band_d(adm_dwt_band_t_d *band, char *data_top, size_t buf_sz_one)

cdef extern from "../../../libvmaf/src/mem.c":
    void *aligned_malloc(size_t size, size_t alignment)
    void aligned_free(void *ptr)

cdef extern from "../../../libvmaf/src/feature/offset.c":
    int offset_image_s(float *buf, float off, int width, int height, int stride)  # TODO: find out why need this _s - doesn't seem to be called but if not, symbol not found error

def adm_dwt2_cy(np.ndarray[np.float64_t, ndim=2, mode='c'] a):

    cdef np.ndarray[np.float64_t, ndim=2, mode='c'] a_buf = np.ascontiguousarray(a, dtype=np.float64)
    cdef double *aa = <double*> a_buf.data  # aa: curr_ref_scale

    cdef int h = len(a)
    cdef int w = len(a[0])
    cdef int h_new, w_new

    cdef double *data_buf = NULL
    cdef char *data_top

    cdef char *ind_buf_y = NULL
    cdef char *ind_buf_x = NULL
    cdef char *buf_y_orig = NULL
    cdef char *buf_x_orig = NULL
    cdef int *ind_y[4]
    cdef int *ind_x[4]

    cdef adm_dwt_band_t_d aa_dwt2  # aa_dwt2: ref_dwt2

    cdef int curr_ref_stride = w * sizeof(double)
    cdef int buf_stride = ALIGN_CEIL(((w + 1) // 2) * sizeof(double))
    cdef size_t buf_sz_one = <size_t> buf_stride * ((h + 1) // 2)

    cdef int ind_size_y = ALIGN_CEIL(((h + 1) // 2) * sizeof(int))
    cdef int ind_size_x = ALIGN_CEIL(((w + 1) // 2) * sizeof(int))

    # == # must use calloc to initialize mem to 0: adm_dwt2_s doesn't touch every cell for small w and h ==
    # data_buf = <double *> aligned_malloc(buf_sz_one * 4, MAX_ALIGN)
    data_buf = <double *> calloc(buf_sz_one * 4, 1)
    if not data_buf:
        free(data_buf)
        aligned_free(buf_y_orig)
        aligned_free(buf_x_orig)
        raise MemoryError
    data_top = <char *>data_buf
    data_top = init_dwt_band_d(&aa_dwt2, data_top, buf_sz_one)

    buf_y_orig = <char *> aligned_malloc(ind_size_y * 4, MAX_ALIGN)
    if not buf_y_orig:
        free(data_buf)
        aligned_free(buf_y_orig)
        aligned_free(buf_x_orig)
        raise MemoryError
    ind_buf_y = <char *>buf_y_orig
    ind_y[0] = <int *> ind_buf_y; ind_buf_y += ind_size_y
    ind_y[1] = <int *> ind_buf_y; ind_buf_y += ind_size_y
    ind_y[2] = <int *> ind_buf_y; ind_buf_y += ind_size_y
    ind_y[3] = <int *> ind_buf_y; ind_buf_y += ind_size_y

    buf_x_orig = <char *> aligned_malloc(ind_size_x * 4, MAX_ALIGN)
    if not buf_x_orig:
        free(data_buf)
        aligned_free(buf_y_orig)
        aligned_free(buf_x_orig)
        raise MemoryError
    ind_buf_x = <char *>buf_x_orig
    ind_x[0] = <int *> ind_buf_x; ind_buf_x += ind_size_x
    ind_x[1] = <int *> ind_buf_x; ind_buf_x += ind_size_x
    ind_x[2] = <int *> ind_buf_x; ind_buf_x += ind_size_x
    ind_x[3] = <int *> ind_buf_x; ind_buf_x += ind_size_x

    cdef np.ndarray[np.float64_t, ndim=2, mode='c'] out_a, out_v, out_h, out_d

    try:
        dwt2_src_indices_filt_s(ind_y, ind_x, w, h)
        adm_dwt2_d(aa, &aa_dwt2, ind_y, ind_x, w, h, curr_ref_stride, buf_stride)

        w_new = (w + 1) // 2
        h_new = (h + 1) // 2

        w_new_strided = ALIGN_CEIL(w_new * sizeof(double)) // sizeof(double)

        # # # ====== debug ======
        # print("h={}, w={}, aa[0]={}, aa[1]={}, aa[2]={}".format(h, w, aa[0], aa[1], aa[2]))
        # print("sizeof(double)={}".format(sizeof(double)))
        # print("curr_ref_stride={}, buf_stride={}, buf_sz_one={}".format(curr_ref_stride, buf_stride, buf_sz_one))
        # print("ind_size_y={}, ind_size_x={}".format(ind_size_y, ind_size_x))
        # print("h_new={}, w_new={}".format(h_new, w_new))
        # print("np.std(aa_band.band_a)={}".format(np.std(np.asarray(<np.float64_t[:h_new, :w_new]> aa_dwt2.band_a))))
        # print("np.std(aa_band.band_v)={}".format(np.std(np.asarray(<np.float64_t[:h_new, :w_new]> aa_dwt2.band_v))))
        # print("np.std(aa_band.band_h)={}".format(np.std(np.asarray(<np.float64_t[:h_new, :w_new]> aa_dwt2.band_h))))
        # print("np.std(aa_band.band_d)={}".format(np.std(np.asarray(<np.float64_t[:h_new, :w_new]> aa_dwt2.band_d))))
        # for i in range(99):  # for 11 x 9
        #     sys.stdout.write("{}\t".format(aa_dwt2.band_a[i]))
        #     if i%10 == 0:
        #         sys.stdout.write("\n")
        # sys.stdout.write("\n")

        out_a = np.empty((h_new, w_new)).astype(np.float64)
        out_v = np.empty((h_new, w_new)).astype(np.float64)
        out_h = np.empty((h_new, w_new)).astype(np.float64)
        out_d = np.empty((h_new, w_new)).astype(np.float64)

        out_a[...] = np.asarray(<np.float64_t[:h_new, :w_new_strided]> aa_dwt2.band_a)[:, :w_new]
        out_v[...] = np.asarray(<np.float64_t[:h_new, :w_new_strided]> aa_dwt2.band_v)[:, :w_new]
        out_h[...] = np.asarray(<np.float64_t[:h_new, :w_new_strided]> aa_dwt2.band_h)[:, :w_new]
        out_d[...] = np.asarray(<np.float64_t[:h_new, :w_new_strided]> aa_dwt2.band_d)[:, :w_new]

    finally:
        free(data_buf)
        aligned_free(buf_y_orig)
        aligned_free(buf_x_orig)

    return out_a, out_v, out_h, out_d



from vmaf.config import VmafConfig

dataset_name = 'test_image'
yuv_fmt = 'yuv444p'

ref_videos = [
 {'content_id': 0,
  'content_name': '100007',
  'height': 321,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100007.yuv'),
  'width': 481},
 {'content_id': 1,
  'content_name': '100039',
  'height': 321,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100039.yuv'),
  'width': 481},
 {'content_id': 2,
  'content_name': '100075',
  'height': 321,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100075.yuv'),
  'width': 481},
 {'content_id': 4,
  'content_name': '100098',
  'height': 321,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100098.yuv'),
  'width': 481},
 {'content_id': 5,
  'content_name': '100099',
  'height': 321,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100099.yuv'),
  'width': 481},
 {'content_id': 6,
  'content_name': '10081',
  'height': 321,
  'path': VmafConfig.test_resource_path('test_image_yuv', '10081.yuv'),
  'width': 481},
 {'content_id': 7,
  'content_name': '101027',
  'height': 321,
  'path': VmafConfig.test_resource_path('test_image_yuv', '101027.yuv'),
  'width': 481},
 {'content_id': 12,
  'content_name': '102062',
  'height': 321,
  'path': VmafConfig.test_resource_path('test_image_yuv', '102062.yuv'),
  'width': 481},
]

dis_videos = [
 {'asset_id': 0,
  'content_id': 0,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100007.yuv'), 'mos': 2.5},
 {'asset_id': 1,
  'content_id': 1,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100039.yuv'), 'mos': 3.9},
 {'asset_id': 2,
  'content_id': 2,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100075.yuv'), 'mos': 5.0},
 {'asset_id': 4,
  'content_id': 4,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100098.yuv'), 'mos': 4.2},
 {'asset_id': 5,
  'content_id': 5,
  'path': VmafConfig.test_resource_path('test_image_yuv', '100099.yuv'), 'mos': 4.1},
 {'asset_id': 6,
  'content_id': 6,
  'path': VmafConfig.test_resource_path('test_image_yuv', '10081.yuv'), 'mos': 4.6},
 {'asset_id': 7,
  'content_id': 7,
  'path': VmafConfig.test_resource_path('test_image_yuv', '101027.yuv'), 'mos': 4.9},
 {'asset_id': 12,
  'content_id': 12,
  'path': VmafConfig.test_resource_path('test_image_yuv', '102062.yuv'), 'mos': 3.3},
]



#!/usr/bin/env python3

import os
import sys
import re

import json

from vmaf.core.result import Result

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


def print_usage():
    print("usage: python " + os.path.basename(sys.argv[0]) + " input_file\n")
    print("input_file contains a list of files for assembly (can be xml or json)")


class FileAssembler:

    SUPPORTED_FILE_TYPES = ['xml', 'json']

    def __init__(self, to_assemble_input):
        self.to_assemble_input = to_assemble_input

    @staticmethod
    def create_assembly_file_list(to_assemble_input):

        to_assemble_list = []
        if isinstance(to_assemble_input, list):
            to_assemble_list = to_assemble_input
        else:
            with open(to_assemble_input, "rt") as input_file:
                for line in input_file.readlines():

                    # match comment
                    mo = re.match(r"^#", line)
                    if mo:
                        print("Skip commented line: {}".format(line))
                        continue

                    # match whitespace
                    mo = re.match(r"[\s]+", line)
                    if mo:
                        continue

                    mo = re.match(r"([\S]+)", line)
                    if not mo:
                        print("Invalid file: {}".format(line))
                        print_usage()
                        return 1

                    to_assemble_list.append(line.strip())

        return to_assemble_list

    def _create_result_list(self, to_assemble_list):
        pass

    def assemble(self):
        """
        Main file assembly logic
        """
        to_assemble_list = self.create_assembly_file_list(self.to_assemble_input)
        self._assert(to_assemble_list)
        results = self._create_result_list(to_assemble_list)
        combined_result = Result.combine_result(results)

        return combined_result

    def _assert(self, to_assemble_list):
        """
        Perform necessary assertions before parsing any of the files.
        """

        # check that the number of files is greater than 0
        assert len(to_assemble_list) > 0
        # check that the file formats match
        assemble_format_list = [os.path.splitext(f)[1].split(".")[1] for f in to_assemble_list]
        assert len(set(assemble_format_list)) == 1, "The file formats for assembly do not much."
        # check that the file format is supported for assembly
        assert assemble_format_list[0] in self.SUPPORTED_FILE_TYPES, \
            "The assembly format is not consistent, use any of {fmts}".format(fmts=str(self.SUPPORTED_FILE_TYPES))


class XmlAssembler(FileAssembler):

    @staticmethod
    def _parse_files(to_assemble_list):

        to_assemble_xml_strings = []
        for to_assemble_xml in to_assemble_list:
            with open(to_assemble_xml, 'r') as f:
                to_assemble_xml_strings.append(f.read())

        return to_assemble_xml_strings

    def _create_result_list(self, to_assemble_list):

        to_assemble_xml_strings = self._parse_files(to_assemble_list)

        results = []
        for to_assemble_xml_string in to_assemble_xml_strings:
            results.append(Result.from_xml(to_assemble_xml_string))

        return results


class JsonAssembler(FileAssembler):

    @staticmethod
    def _parse_files(to_assemble_list):

        to_assemble_json_strings = []
        for json_file in to_assemble_list:
            with open(json_file, 'r') as f:
                to_assemble_json_strings.append(json.load(f))

        return to_assemble_json_strings

    def _create_result_list(self, to_assemble_list):

        to_assemble_jsons = self._parse_files(to_assemble_list)

        results = []
        for to_assemble_json in to_assemble_jsons:
            to_assemble_json_string = json.dumps(to_assemble_json)
            results.append(Result.from_json(to_assemble_json_string))

        return results


def main():

    if len(sys.argv) != 2:
        print_usage()
        return 2

    files_to_assemble_list = sys.argv[1]

    desired_file_list = FileAssembler.create_assembly_file_list(files_to_assemble_list)

    if ".xml" in desired_file_list[0]:
        xml_output = XmlAssembler(files_to_assemble_list).assemble().to_xml()
        print(xml_output)
    elif ".json" in desired_file_list[0]:
        json_output = JsonAssembler(files_to_assemble_list).assemble().to_json()
        print(json_output)
    else:
        print_usage()
        return 2


if __name__ == "__main__":
    ret = main()
    exit(ret)



#!/usr/bin/env python3

import matplotlib
matplotlib.use('Agg')

import sys
import os

import numpy as np

from vmaf.config import VmafConfig, DisplayConfig
from vmaf.core.asset import Asset
from vmaf.core.quality_runner import VmafQualityRunner
from vmaf.tools.misc import get_file_name_without_extension, get_cmd_option, \
    cmd_option_exists
from vmaf.tools.stats import ListStats

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

FMTS = ['yuv420p', 'yuv422p', 'yuv444p', 
        'yuv420p10le', 'yuv422p10le', 'yuv444p10le',
        'yuv420p12le', 'yuv422p12le', 'yuv444p12le',
        'yuv420p16le', 'yuv422p16le', 'yuv444p16le',
        ]
OUT_FMTS = ['text (default)', 'xml', 'json']
POOL_METHODS = ['mean', 'harmonic_mean', 'min', 'median', 'perc5', 'perc10', 'perc20']


def print_usage():
    print("usage: " + os.path.basename(sys.argv[0]) \
          + " fmt width height ref_path dis_path [--model model_path] [--out-fmt out_fmt] " \
            "[--phone-model] [--ci] [--save-plot plot_dir]\n")
    print("fmt:\n\t" + "\n\t".join(FMTS) + "\n")
    print("out_fmt:\n\t" + "\n\t".join(OUT_FMTS) + "\n")


def main():
    if len(sys.argv) < 6:
        print_usage()
        return 2

    try:
        fmt = sys.argv[1]
        width = int(sys.argv[2])
        height = int(sys.argv[3])
        ref_file = sys.argv[4]
        dis_file = sys.argv[5]
    except ValueError:
        print_usage()
        return 2

    if width < 0 or height < 0:
        print("width and height must be non-negative, but are {w} and {h}".format(w=width, h=height))
        print_usage()
        return 2

    if fmt not in FMTS:
        print_usage()
        return 2

    model_path = get_cmd_option(sys.argv, 6, len(sys.argv), '--model')

    out_fmt = get_cmd_option(sys.argv, 6, len(sys.argv), '--out-fmt')
    if not (out_fmt is None
            or out_fmt == 'xml'
            or out_fmt == 'json'
            or out_fmt == 'text'):
        print_usage()
        return 2

    pool_method = get_cmd_option(sys.argv, 6, len(sys.argv), '--pool')
    if not (pool_method is None
            or pool_method in POOL_METHODS):
        print('--pool can only have option among {}'.format(', '.join(POOL_METHODS)))
        return 2

    show_local_explanation = cmd_option_exists(sys.argv, 6, len(sys.argv), '--local-explain')

    phone_model = cmd_option_exists(sys.argv, 6, len(sys.argv), '--phone-model')

    enable_conf_interval = cmd_option_exists(sys.argv, 6, len(sys.argv), '--ci')

    save_plot_dir = get_cmd_option(sys.argv, 6, len(sys.argv), '--save-plot')

    if show_local_explanation and enable_conf_interval:
        print('cannot set both --local-explain and --ci flags')
        return 2

    asset = Asset(dataset="cmd",
                  content_id=abs(hash(get_file_name_without_extension(ref_file))) % (10 ** 16),
                  asset_id=abs(hash(get_file_name_without_extension(ref_file))) % (10 ** 16),
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_file,
                  dis_path=dis_file,
                  asset_dict={'width':width, 'height':height, 'yuv_type':fmt}
                  )
    assets = [asset]

    if show_local_explanation:
        from vmaf.core.quality_runner_extra import VmafQualityRunnerWithLocalExplainer
        runner_class = VmafQualityRunnerWithLocalExplainer
    elif enable_conf_interval:
        from vmaf.core.quality_runner import BootstrapVmafQualityRunner
        runner_class = BootstrapVmafQualityRunner
    else:
        runner_class = VmafQualityRunner

    if model_path is None:
        optional_dict = None
    else:
        optional_dict = {'model_filepath':model_path}

    if phone_model:
        if optional_dict is None:
            optional_dict = {}
        optional_dict['enable_transform_score'] = True

    runner = runner_class(
        assets, None, fifo_mode=True,
        delete_workdir=True,
        result_store=None,
        optional_dict=optional_dict,
        optional_dict2=None,
    )

    # run
    runner.run()
    result = runner.results[0]

    # pooling
    if pool_method == 'harmonic_mean':
        result.set_score_aggregate_method(ListStats.harmonic_mean)
    elif pool_method == 'min':
        result.set_score_aggregate_method(np.min)
    elif pool_method == 'median':
        result.set_score_aggregate_method(np.median)
    elif pool_method == 'perc5':
        result.set_score_aggregate_method(ListStats.perc5)
    elif pool_method == 'perc10':
        result.set_score_aggregate_method(ListStats.perc10)
    elif pool_method == 'perc20':
        result.set_score_aggregate_method(ListStats.perc20)
    else: # None or 'mean'
        pass

    # output
    if out_fmt == 'xml':
        print(result.to_xml())
    elif out_fmt == 'json':
        print(result.to_json())
    else: # None or 'text'
        print(str(result))

    # local explanation
    if show_local_explanation:
        runner.show_local_explanations([result])

        if save_plot_dir is None:
            DisplayConfig.show()
        else:
            DisplayConfig.show(write_to_dir=save_plot_dir)

    return 0


if __name__ == "__main__":
    ret = main()
    exit(ret)



dataset_name = 'example'

from vmaf.config import VmafConfig

ref_videos = [
    {'content_id': 0, 'path': VmafConfig.resource_path('yuv', 'checkerboard_1920_1080_10_3_0_0.yuv'), 'width': 1920, 'height': 1080, 'yuv_fmt': 'yuv420p'},

    {'content_id': 1, 'path': VmafConfig.resource_path('yuv', 'flat_1280_720_0.yuv'), 'width': 1280, 'height': 720, 'yuv_fmt': 'yuv420p10le'},
]

dis_videos = [
    {'content_id': 0, 'asset_id': 0, 'groundtruth': 100, 'path': VmafConfig.resource_path('yuv', 'checkerboard_1920_1080_10_3_0_0.yuv')}, # ref
    {'content_id': 0, 'asset_id': 1, 'groundtruth': 50, 'path': VmafConfig.resource_path('yuv', 'checkerboard_1920_1080_10_3_1_0.yuv')},

    {'content_id': 1, 'asset_id': 2, 'groundtruth': 100, 'path': VmafConfig.resource_path('yuv', 'flat_1280_720_0.yuv')}, # ref
    {'content_id': 1, 'asset_id': 3, 'groundtruth': 80, 'path': VmafConfig.resource_path('yuv', 'flat_1280_720_10.yuv')},
]



from collections import OrderedDict
import json
import re
from typing import Optional, Callable

import numpy as np

from vmaf.tools.misc import get_file_name_with_extension
from vmaf.core.asset import Asset

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class BasicResult(object):
    """
    Has some basic functions, but don't need asset or executor_id. To be used by
    FeatureAssembler, which is not an Executor.
    """
    def __init__(self, asset, result_dict):
        self.asset = asset
        self.result_dict = result_dict
        self.score_aggregate_method = np.mean

    def set_score_aggregate_method(self, score_aggregate_method: Optional[Callable]):
        if score_aggregate_method is not None:
            self.score_aggregate_method = score_aggregate_method
        else:
            self.score_aggregate_method = np.mean

    # make access dictionary-like, i.e. can do: result['vif_score']
    def __getitem__(self, key):
        return self.get_result(key)

    def get_result(self, key):
        try:
            return self.result_dict[key]
        except KeyError as e:
            return self._try_get_aggregate_score(key, e)

    @staticmethod
    def get_scores_key_from_score_key(score_key):
        # e.g. 'VMAF_scores'
        return score_key + 's'

    def _try_get_aggregate_score(self, key, error):
        """
        Get aggregate score from list of scores. Must follow the convention
        that if the aggregate score uses key '*_score', then there must be
        a corresponding list of scores that uses key '*_scores'. For example,
        if the key is 'VMAF_score', there must exist a corresponding key
        'VMAF_scores'.
        The list of scores (when present) should also follow the convention
        that IF multiple models are present (e.g. in bootstrapping),
        the first dimension will be the model and the
        second dimension the video frames. This means that a 2D result array
        should be created ONLY when multiple models are present.
        """
        if re.search(r"_score$", key):
            scores_key = self.get_scores_key_from_score_key(key)
            if scores_key in self.result_dict:
                scores = self.result_dict[scores_key]
                # if scores are in a list, wrap in an array to be consistent
                if type(scores) is list:
                    scores = np.asarray(scores)
                # dimension assertion: scores should be either 1-D (one prediction per frame) or 2-D (single/multiple predictions per frame)
                assert scores.ndim <= 2, 'Per frame score aggregation is not well-defined; scores cannot be saved in a N-D array with N > 2.'

                # check if there are more than one models (first dimension)
                if scores.ndim == 2:
                    # check that there are > 1 models present
                    assert scores.shape[0] > 1, '# models is <=1, but a 2D result array (models x frames) was used.'
                    # check that there are >= 1 frames predicted
                    assert scores.shape[1] >= 1, '# predicted frames is < 1.'
                    # apply score aggregation on each individual model
                    # a scores "piece" corresponds to a single model's predictions over all frames
                    return [self.score_aggregate_method(scores_piece) for scores_piece in scores]
                else:
                    # just one prediction per frame
                    return self.score_aggregate_method(scores)
        raise KeyError(error)

    def get_ordered_list_scores_key(self):
        # e.g. ['VMAF_scores', 'VMAF_vif_scores']
        # filter out scores that are > 1D (e.g. when having multiple models in bootstrapping)
        list_scores_key = filter(lambda key: re.search(r"_scores$", key) and np.asarray(self.result_dict[key]).ndim == 1,
                                 self.result_dict.keys())
        list_scores_key = sorted(list_scores_key)
        return list_scores_key

    def get_ordered_list_multimodel_scores_key(self):
        # e.g. ['BOOTSTRAP_VMAF_all_models_scores']
        # select only scores that are > 1D (e.g. when having multiple models in bootstrapping)
        list_scores_key = filter(lambda key: re.search(r"_scores$", key) and np.asarray(self.result_dict[key]).ndim > 1,
                                 self.result_dict.keys())
        list_scores_key = sorted(list_scores_key)
        return list_scores_key

    def get_ordered_list_score_key(self):
        # e.g. ['VMAF_score', 'VMAF_vif_score']
        list_scores_key = self.get_ordered_list_scores_key()
        return list(map(lambda scores_key: scores_key[:-1], list_scores_key))

    def get_ordered_list_multimodel_score_key(self):
        # e.g. ['BOOTSTRAP_VMAF_all_models_score']
        list_scores_key = self.get_ordered_list_multimodel_scores_key()
        return list(map(lambda scores_key: scores_key[:-1], list_scores_key))

    def _get_scores_str(self, unit_name='Frame'):
        list_scores_key = self.get_ordered_list_scores_key()
        list_score_key = self.get_ordered_list_score_key()
        list_scores = list(map(lambda key: self.result_dict[key], list_scores_key))
        str_perframe = "\n".join(
            list(map(
                lambda tframe_scores: "{unit} {num}: ".format(
                    unit=unit_name, num=tframe_scores[0]) + (
                ", ".join(
                    list(map(
                        lambda tscore: "{score_key}:{score:.6f}".format(score_key=tscore[0], score=tscore[1]),
                        zip(list_score_key, tframe_scores[1])))
                )),
                enumerate(zip(*list_scores))
            ))
        )
        str_perframe += '\n'
        return str_perframe

    def _get_aggregate_score_str(self):
        list_score_key = self.get_ordered_list_score_key()
        str_aggregate = "Aggregate ({}): ".format(self.score_aggregate_method.__name__) + (", ".join(
            list(map(
                lambda tscore: "{score_key}:{score:.6f}".format(score_key=tscore[0], score=tscore[1]),
                zip(
                    list_score_key, list(map(
                        lambda score_key: self[score_key],
                        list_score_key)
                ))
            ))
        ))
        return str_aggregate

    @staticmethod
    def scores_key_wildcard_match(result_dict, scores_key, excluded_scores_keys=None):
        """
        Find all matching score keys from a result dictionary. When a list of keys is provided via excluded_scores_keys,
        these particular keys will be ommited from the query result.
        >>> BasicResult.scores_key_wildcard_match({'VMAF_integer_feature_vif_scale0_egl_1_scores': [0.983708]}, 'VMAF_integer_feature_vif_scale0_scores')
        'VMAF_integer_feature_vif_scale0_egl_1_scores'
        >>> BasicResult.scores_key_wildcard_match({'VMAF_integer_feature_vif_scale0_egl_1_scores': [0.983708]}, 'VMAF_integer_feature_vif_scale1_scores')
        Traceback (most recent call last):
        ...
        KeyError: 'no key matches VMAF_integer_feature_vif_scale1_scores'
        >>> d = {'VMAF_feature_adm_den_scale0_scores': [111.207703], \
                 'VMAF_feature_adm_den_scale1_scores': [109.423508], \
                 'VMAF_feature_adm_den_scale2_scores': [196.427551], \
                 'VMAF_feature_adm_den_scale3_scores': [328.864075], \
                 'VMAF_feature_adm_den_scores': [745.922836], \
                 'VMAF_feature_adm_num_scale0_scores': [107.988159], \
                 'VMAF_feature_adm_num_scale1_scores': [96.965546], \
                 'VMAF_feature_adm_num_scale2_scores': [178.08934], \
                 'VMAF_feature_adm_num_scale3_scores': [317.582733], \
                 'VMAF_feature_adm_num_scores': [700.625778], \
                 'VMAF_feature_adm_scores': [0.939274], \
                 'VMAF_feature_anpsnr_scores': [41.921087], \
                 'VMAF_feature_ansnr_scores': [30.230282], \
                 'VMAF_feature_motion2_scores': [0.0], \
                 'VMAF_feature_motion_scores': [0.0], \
                 'VMAF_feature_vif_den_scale0_scores': [30884050.0], \
                 'VMAF_feature_vif_den_scale1_scores': [7006361.0], \
                 'VMAF_feature_vif_den_scale2_scores': [1696758.0], \
                 'VMAF_feature_vif_den_scale3_scores': [429003.59375], \
                 'VMAF_feature_vif_den_scores': [40016172.59375], \
                 'VMAF_feature_vif_num_scale0_scores': [18583060.0], \
                 'VMAF_feature_vif_num_scale1_scores': [5836731.5], \
                 'VMAF_feature_vif_num_scale2_scores': [1545453.125], \
                 'VMAF_feature_vif_num_scale3_scores': [408037.5], \
                 'VMAF_feature_vif_num_scores': [26373282.125], \
                 'VMAF_feature_vif_scores': [0.659066]}
        >>> BasicResult.scores_key_wildcard_match(d, 'VMAF_feature_adm_num_scores')
        'VMAF_feature_adm_num_scores'
        >>> BasicResult.scores_key_wildcard_match(d, 'VMAF_feature_adm_num_scale_scores')
        Traceback (most recent call last):
        ...
        KeyError: "more than one keys matches VMAF_feature_adm_num_scale_scores: ['VMAF_feature_adm_num_scale0_scores', 'VMAF_feature_adm_num_scale1_scores', 'VMAF_feature_adm_num_scale2_scores', 'VMAF_feature_adm_num_scale3_scores']"
        >>> e = {'VMAF_feature_adm_den_egl_1_scores': [290.148182], \
                 'VMAF_feature_adm_den_scale0_egl_1_scores': [36.143059], \
                 'VMAF_feature_adm_den_scale1_egl_1_scores': [56.709286], \
                 'VMAF_feature_adm_den_scale2_egl_1_scores': [83.719971], \
                 'VMAF_feature_adm_den_scale3_egl_1_scores': [113.575867], \
                 'VMAF_feature_adm_num_egl_1_scores': [277.796806], \
                 'VMAF_feature_adm_num_scale0_egl_1_scores': [35.399879], \
                 'VMAF_feature_adm_num_scale1_egl_1_scores': [54.699654], \
                 'VMAF_feature_adm_num_scale2_egl_1_scores': [80.061874], \
                 'VMAF_feature_adm_num_scale3_egl_1_scores': [107.635399], \
                 'VMAF_feature_adm_scale0_egl_1_scores': [0.979438], \
                 'VMAF_feature_anpsnr_scores': [24.463257], \
                 'VMAF_feature_ansnr_scores': [11.619336], \
                 'VMAF_feature_motion2_scores': [0.0], \
                 'VMAF_feature_motion_scores': [0.0], \
                 'VMAF_feature_vif_den_egl_1_scores': [584014.634277], \
                 'VMAF_feature_vif_den_scale0_egl_1_scores': [452763.03125], \
                 'VMAF_feature_vif_den_scale1_egl_1_scores': [100037.859375], \
                 'VMAF_feature_vif_den_scale2_egl_1_scores': [24712.830078], \
                 'VMAF_feature_vif_den_scale3_egl_1_scores': [6500.913574], \
                 'VMAF_feature_vif_num_egl_1_scores': [576352.736328], \
                 'VMAF_feature_vif_num_scale0_egl_1_scores': [445400.1875], \
                 'VMAF_feature_vif_num_scale1_egl_1_scores': [99781.773438], \
                 'VMAF_feature_vif_num_scale2_egl_1_scores': [24675.099609], \
                 'VMAF_feature_vif_num_scale3_egl_1_scores': [6495.675781], \
                 'VMAF_feature_vif_scale0_egl_1_scores': [0.983738]}
        >>> BasicResult.scores_key_wildcard_match(e, 'VMAF_feature_adm_num_scores')
        'VMAF_feature_adm_num_egl_1_scores'
        >>> f = {'VMAF_integer_feature_vif_scale0_egl_1_scores': [111.207703], \
                 'VMAF_integer_feature_vif_scale0_eg_1_scores': [109.423508]}
        >>> BasicResult.scores_key_wildcard_match(f, 'VMAF_integer_feature_vif_scale0_scores')
        'VMAF_integer_feature_vif_scale0_eg_1_scores'
        >>> BasicResult.scores_key_wildcard_match(f, 'VMAF_integer_feature_vif_scale0_scores', excluded_scores_keys=[])
        'VMAF_integer_feature_vif_scale0_eg_1_scores'
        >>> BasicResult.scores_key_wildcard_match(f, 'VMAF_integer_feature_vif_scale0_scores', excluded_scores_keys=['VMAF_integer_feature_vif_scale0_eg_1_scores'])
        'VMAF_integer_feature_vif_scale0_egl_1_scores'
        >>> BasicResult.scores_key_wildcard_match(f, 'VMAF_integer_feature_vif_scale0_scores', excluded_scores_keys=['VMAF_integer_feature_vif_scale0_eg_1_scores', 'VMAF_integer_feature_vif_scale0_egl_1_scores'])
        Traceback (most recent call last):
        ...
        KeyError: 'no key matches VMAF_integer_feature_vif_scale0_scores'
        """

        # first look for exact match
        result_keys_sorted = [key for key in sorted(result_dict.keys()) if (excluded_scores_keys is not None and
                                                                            key not in excluded_scores_keys) or
                              excluded_scores_keys is None]
        for result_key in result_keys_sorted:
            if result_key == scores_key:
                return result_key

        # then look for wildcard match
        matched_result_keys = []
        for result_key in result_keys_sorted:
            if result_key.startswith(scores_key[:-len('_scores')]):
                matched_result_keys.append(result_key)
        if len(matched_result_keys) == 0:
            raise KeyError(f"no key matches {scores_key}")
        elif len(matched_result_keys) == 1:
            return matched_result_keys[0]
        else:
            # look for shortest match
            strlens = [len(s) for s in matched_result_keys]
            argmin_strlen = np.concatenate(np.where(strlens == np.min(strlens))).tolist()
            if len(argmin_strlen) == 1:
                return matched_result_keys[argmin_strlen[0]]
            else:
                raise KeyError(f"more than one keys matches {scores_key}: {matched_result_keys}")


class Result(BasicResult):
    """
    Dictionary-like object that stores read-only result generated on an Asset
    by a Executor.
    """
    DATAFRAME_COLUMNS = (
        'dataset',
        'content_id',
        'asset_id',
        'ref_name',
        'dis_name',
        'asset',
        'executor_id',
        'scores_key',
        'scores' # one score per unit - frame, chunk or else
    )

    def __init__(self, asset, executor_id, result_dict):
        super(Result, self).__init__(asset, result_dict)
        self.executor_id = executor_id

    def __eq__(self, other):
        if self.asset != other.asset:
            return False
        if self.executor_id != other.executor_id:
            return False
        list_scores_key = self.get_ordered_list_scores_key()
        for scores_key in list_scores_key:
            if self.result_dict[scores_key] != other.result_dict[scores_key]:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    # TODO: add __repr__, __hash__

    def __str__(self):
        return self.to_string()

    @staticmethod
    def get_unique_from_dataframe(df, scores_key, column):
        """
        Convenience method to access dataframe. Do two things 1) make assertion
        that only one row correspond to the scores_key, 2) retrive a column
        from that row
        :param df:
        :param scores_key:
        :param column:
        :return:
        """
        _df = df.loc[df['scores_key'] == scores_key]
        assert len(_df) == 1
        return _df.iloc[0][column]

    def to_string(self):
        """Example:
        Asset: {"asset_dict": {"height": 1080, "width": 1920}, "asset_id": 0, "content_id": 0, "dataset": "test",
            "dis_path": ".../vmaf/python/test/resource/yuv/checkerboard_1920_1080_10_3_1_0.yuv",
            "ref_path": ".../vmaf/python/test/resource/yuv/checkerboard_1920_1080_10_3_0_0.yuv",
            "workdir": ".../vmaf/workspace/workdir/d26050af-bd92-46a7-8519-7482306aa7fe"}
        Executor: SSIM_V1.0
        Result:
        Frame 0: SSIM_feature_ssim_c_score:0.997404, SSIM_feature_ssim_l_score:0.965512, SSIM_feature_ssim_s_score:0.935803, SSIM_score:0.901161
        Frame 1: SSIM_feature_ssim_c_score:0.997404, SSIM_feature_ssim_l_score:0.965512, SSIM_feature_ssim_s_score:0.935803, SSIM_score:0.901160
        Frame 2: SSIM_feature_ssim_c_score:0.997404, SSIM_feature_ssim_l_score:0.965514, SSIM_feature_ssim_s_score:0.935804, SSIM_score:0.901163
        Aggregate: SSIM_feature_ssim_c_score:0.997404, SSIM_feature_ssim_l_score:0.965513, SSIM_feature_ssim_s_score:0.935803, SSIM_score:0.901161
        """
        s = ""
        s += 'Asset: {}\n'.format( # unlike repr(asset), print workdir
            self.asset.to_full_repr())
        s += 'Executor: {}\n'.format(self.executor_id)
        s += 'Result:\n'
        s += self._get_scores_str()
        s += self._get_aggregate_score_str()
        return s

    def to_xml(self):
        """Example:
        <?xml version="1.0" ?>
        <result executorId="SSIM_V1.0">
          <asset identifier="test_0_0_checkerboard_1920_1080_10_3_0_0_1920x1080_vs_checkerboard_1920_1080_10_3_1_0_1920x1080_q_1920x1080"/>
          <frames>
            <frame SSIM_feature_ssim_c_score="0.997404" SSIM_feature_ssim_l_score="0.965512" SSIM_feature_ssim_s_score="0.935803" SSIM_score="0.901161" frameNum="0"/>
            <frame SSIM_feature_ssim_c_score="0.997404" SSIM_feature_ssim_l_score="0.965512" SSIM_feature_ssim_s_score="0.935803" SSIM_score="0.90116" frameNum="1"/>
            <frame SSIM_feature_ssim_c_score="0.997404" SSIM_feature_ssim_l_score="0.965514" SSIM_feature_ssim_s_score="0.935804" SSIM_score="0.901163" frameNum="2"/>
          </frames>
          <aggregate SSIM_feature_ssim_c_score="0.997404" SSIM_feature_ssim_l_score="0.965512666667" SSIM_feature_ssim_s_score="0.935803333333" SSIM_score="0.901161333333"/>
        </result>
        """

        from xml.etree import ElementTree
        from xml.dom import minidom

        list_scores_key = self.get_ordered_list_scores_key()
        list_score_key = self.get_ordered_list_score_key()
        list_scores = list(map(lambda key: self.result_dict[key], list_scores_key))
        list_aggregate_score = list(map(lambda key: self[key], list_score_key))

        list_multimodel_scores_key = self.get_ordered_list_multimodel_scores_key()
        list_multimodel_score_key = self.get_ordered_list_multimodel_score_key()
        # here we need to transpose, since printing is per frame and not per model
        # we also need to turn the 2D array to a list of lists, for unpacking to work as expected
        list_multimodel_scores = list(map(lambda key: self.result_dict[key].T.tolist(), list_multimodel_scores_key))
        list_aggregate_multimodel_score = list(map(lambda key: self[key], list_multimodel_score_key))

        # append multimodel scores and keys (if any)
        list_scores_key += list_multimodel_scores_key
        list_score_key += list_multimodel_score_key
        list_scores += list_multimodel_scores
        list_aggregate_score += list_aggregate_multimodel_score

        list_scores_reordered = zip(*list_scores)

        def prettify(elem):
            rough_string = ElementTree.tostring(elem, 'utf-8')
            reparsed = minidom.parseString(rough_string)
            return reparsed.toprettyxml(indent="  ")

        top = ElementTree.Element('result')
        top.set('executorId', self.executor_id)

        asset = ElementTree.SubElement(top, 'asset')
        asset.set('identifier', str(self.asset))

        frames = ElementTree.SubElement(top, 'frames')
        for i, list_score in enumerate(list_scores_reordered):
            frame = ElementTree.SubElement(frames, 'frame')
            frame.set('frameNum', str(i))
            for score_key, score in zip(list_score_key, list_score):
                frame.set(score_key, str(score))

        aggregate = ElementTree.SubElement(top, 'aggregate')
        aggregate.set('method', self.score_aggregate_method.__name__)
        for score_key, score in zip(list_score_key, list_aggregate_score):
            aggregate.set(score_key, str(score))

        return prettify(top)

    @classmethod
    def from_xml(cls, xml_string):

        from xml.etree import ElementTree

        top = ElementTree.fromstring(xml_string)
        asset = top.find('asset').get('identifier')
        executor_id = top.get('executorId')

        result_dict = {}

        frames = top.find('frames')

        pred_result_key = None

        for frame_ind, frame in enumerate(frames):
            for score_key, score_value in frame.attrib.items():
                if 'score' in score_key:
                    # same convention of 'score' and 'scores' as in _try_get_aggregate_score
                    if 'feature' in score_key:
                        result_dict.setdefault(Result.get_scores_key_from_score_key(score_key), []).append(np.float64(score_value))
                    else:
                        pred_result_key = Result.get_scores_key_from_score_key(score_key)
                        result_dict.setdefault(pred_result_key, []).append(np.float64(score_value))

        result_dict[pred_result_key] = np.array(result_dict[pred_result_key])

        return Result(asset, executor_id, result_dict)

    @classmethod
    def from_json(cls, json_string):

        json_data = json.loads(json_string)
        executor_id = json_data['executorId']
        asset = json_data['asset']['identifier']

        result_dict = {}

        frames = json_data['frames']

        pred_result_key = None

        for frame_ind, frame in enumerate(frames):
            for score_key, score_value in frame.items():
                if 'score' in score_key:
                    # same convention of 'score' and 'scores' as in _try_get_aggregate_score
                    if 'feature' in score_key:
                        result_dict.setdefault(Result.get_scores_key_from_score_key(score_key), []).append(np.float64(score_value))
                    else:
                        pred_result_key = Result.get_scores_key_from_score_key(score_key)
                        result_dict.setdefault(pred_result_key, []).append(np.float64(score_value))

        result_dict[pred_result_key] = np.array(result_dict[pred_result_key])

        return Result(asset, executor_id, result_dict)

    @classmethod
    def combine_result(cls, results):

        assert len(results) > 0, "Results list is empty."
        executor_ids = [result.executor_id for result in results]
        assert len(set(executor_ids)) == 1, "Executor ids do not match."
        executor_id = executor_ids[0]
        combined_result_dict = OrderedDict()

        sorted_scores_keys = sorted([key for key in results[0].result_dict.keys()], reverse=True)

        # initialize dictionary
        for scores_key in sorted_scores_keys:
            combined_result_dict[scores_key] = []

        for result in results:
            result_dict  = result.result_dict
            # assert if the keys in result_dict match
            assert sorted([key for key in result_dict.keys()], reverse=True) == sorted_scores_keys, "Score keys do not match."
            for scores_key in sorted_scores_keys:
                combined_result_dict[scores_key] += list(result_dict[scores_key])

        pred_result_key = [key for key in sorted_scores_keys if 'feature' not in key][0]
        combined_result_dict[pred_result_key] = np.array(combined_result_dict[pred_result_key])

        return Result('combined_asset', executor_id, combined_result_dict)

    def to_dict(self):
        """Example:
        {
            "executorId": "SSIM_V1.0",
            "asset": {
                "identifier": "test_0_0_checkerboard_1920_1080_10_3_0_0_1920x1080_vs_checkerboard_1920_1080_10_3_1_0_1920x1080_q_1920x1080"
            },
            "frames": [
                {
                    "frameNum": 0,
                    "SSIM_feature_ssim_c_score": 0.997404,
                    "SSIM_feature_ssim_l_score": 0.965512,
                    "SSIM_feature_ssim_s_score": 0.935803,
                    "SSIM_score": 0.901161
                },
                {
                    "frameNum": 1,
                    "SSIM_feature_ssim_c_score": 0.997404,
                    "SSIM_feature_ssim_l_score": 0.965512,
                    "SSIM_feature_ssim_s_score": 0.935803,
                    "SSIM_score": 0.90116
                },
                {
                    "frameNum": 2,
                    "SSIM_feature_ssim_c_score": 0.997404,
                    "SSIM_feature_ssim_l_score": 0.965514,
                    "SSIM_feature_ssim_s_score": 0.935804,
                    "SSIM_score": 0.901163
                }
            ],
            "aggregate": {
                "SSIM_feature_ssim_c_score": 0.99740399999999996,
                "SSIM_feature_ssim_l_score": 0.96551266666666669,
                "SSIM_feature_ssim_s_score": 0.93580333333333332,
                "SSIM_score": 0.90116133333333337
            }
        }
        """
        list_scores_key = self.get_ordered_list_scores_key()
        list_score_key = self.get_ordered_list_score_key()
        list_scores = list(map(lambda key: self.result_dict[key], list_scores_key))
        list_aggregate_score = list(map(lambda key: self[key], list_score_key))

        list_multimodel_scores_key = self.get_ordered_list_multimodel_scores_key()
        list_multimodel_score_key = self.get_ordered_list_multimodel_score_key()
        # here we need to transpose, since printing is per frame and not per model
        # we also need to turn the 2D array to a list of lists, for unpacking to work as expected
        list_multimodel_scores = list(map(lambda key: self.result_dict[key].T.tolist(), list_multimodel_scores_key))
        list_aggregate_multimodel_score = list(map(lambda key: self[key], list_multimodel_score_key))

        # append multimodel scores and keys (if any)
        list_scores_key += list_multimodel_scores_key
        list_score_key += list_multimodel_score_key
        list_scores += list_multimodel_scores
        list_aggregate_score += list_aggregate_multimodel_score

        list_scores_reordered = zip(*list_scores)

        top = OrderedDict()
        top['executorId'] = self.executor_id
        top['asset'] = {'identifier': str(self.asset)}

        top['frames'] = []
        for i, list_score in enumerate(list_scores_reordered):
            frame = OrderedDict()
            frame['frameNum'] = i
            for score_key, score in zip(list_score_key, list_score):
                frame[score_key] = score
            top['frames'].append(frame)
        top['aggregate'] = OrderedDict()

        for score_key, score in zip(list_score_key, list_aggregate_score):
            top['aggregate'][score_key] = score
        top['aggregate']['method'] = self.score_aggregate_method.__name__

        return top

    def to_json(self):
        """
        :return str: JSON representation
        """
        return json.dumps(self.to_dict(), sort_keys=False, indent=4)

    def to_dataframe(self):
        """
        Export to pandas dataframe with columns: dataset, content_id, asset_id,
        ref_name, dis_name, asset, executor_id, scores_key, scores
        Example:
                                                       asset  asset_id  content_id  \
        0  {"asset_dict": {"height": 1080, "width": 1920}...         0           0
        1  {"asset_dict": {"height": 1080, "width": 1920}...         0           0
        2  {"asset_dict": {"height": 1080, "width": 1920}...         0           0
        3  {"asset_dict": {"height": 1080, "width": 1920}...         0           0
        4  {"asset_dict": {"height": 1080, "width": 1920}...         0           0

          dataset                             dis_name executor_id  \
        0    test  checkerboard_1920_1080_10_3_1_0.yuv   VMAF_V0.1
        1    test  checkerboard_1920_1080_10_3_1_0.yuv   VMAF_V0.1
        2    test  checkerboard_1920_1080_10_3_1_0.yuv   VMAF_V0.1
        3    test  checkerboard_1920_1080_10_3_1_0.yuv   VMAF_V0.1
        4    test  checkerboard_1920_1080_10_3_1_0.yuv   VMAF_V0.1

                                      ref_name  \
        0  checkerboard_1920_1080_10_3_0_0.yuv
        1  checkerboard_1920_1080_10_3_0_0.yuv
        2  checkerboard_1920_1080_10_3_0_0.yuv
        3  checkerboard_1920_1080_10_3_0_0.yuv
        4  checkerboard_1920_1080_10_3_0_0.yuv

                                                  scores          scores_key
        0                  [0.798588, 0.84287, 0.800122]     VMAF_adm_scores
        1               [12.420815, 12.41775, 12.416308]   VMAF_ansnr_scores
        2                    [0.0, 18.489031, 18.542355]  VMAF_motion_scores
        3  [42.1117149479, 47.6544689539, 40.6168118533]         VMAF_scores
        4                 [0.156106, 0.156163, 0.156119]     VMAF_vif_scores

        [5 rows x 9 columns]
        :return:
        """
        import pandas as pd
        asset = self.asset
        executor_id = self.executor_id
        list_scores_key = self.get_ordered_list_scores_key()
        list_scores = list(map(lambda key: self.result_dict[key], list_scores_key))

        rows = []
        for scores_key, scores in zip(list_scores_key, list_scores):
            row = [asset.dataset,
                   asset.content_id,
                   asset.asset_id,
                   get_file_name_with_extension(asset.ref_path),
                   get_file_name_with_extension(asset.dis_path),
                   repr(asset),
                   executor_id,
                   scores_key,
                   scores]
            rows.append(row)

        # zip rows into a dict, and wrap with df
        df = pd.DataFrame(dict(zip(self.DATAFRAME_COLUMNS, zip(*rows))))

        return df

    @classmethod
    def from_dataframe(cls, df, AssetClass=Asset):

        # first, make sure the df conform to the format for a single asset
        cls._assert_asset_dataframe(df)

        asset_repr = df.iloc[0]['asset']
        asset = AssetClass.from_repr(asset_repr)

        executor_id = df.iloc[0]['executor_id']

        result_dict = {}
        for _, row in df.iterrows():
            result_dict[row['scores_key']] = row['scores']

        return Result(asset, executor_id, result_dict)

    @classmethod
    def _assert_asset_dataframe(cls, df):
        """
        Make sure that the input dataframe conforms w.r.t. an asset
        :param df:
        :return:
        """
        # check columns
        for col in list(df.columns.values):
            assert col in cls.DATAFRAME_COLUMNS

        # all rows should have the same dataset, content_id, asset_id, ref_name,
        # dis_name, asset and executor_id
        assert len(set(df['dataset'].tolist())) == 1
        assert len(set(df['content_id'].tolist())) == 1
        assert len(set(df['asset_id'].tolist())) == 1
        assert len(set(df['ref_name'].tolist())) == 1
        assert len(set(df['dis_name'].tolist())) == 1
        assert len(set(df['asset'].tolist())) == 1
        assert len(set(df['executor_id'].tolist())) == 1

        # each scores key must have one single row
        assert len(df) == len(set(df['scores_key'].tolist()))

        # all scores should have equal length
        assert len(set(map(lambda x:len(x), df['scores'].tolist()))) == 1


class RawResult(object):
    """
    Other than sharing the name 'Result' and have same initialization interface,
    RawResult is very different from vmaf.core.result.Result class -- it won't be
    stored using ResultStore, neither it has aggregation method like
    result['vmaf_score'] (which calls _try_get_aggregate_score()).
    """

    def __init__(self, asset, executor_id, result_dict):
        # same interface as Result
        self.asset = asset
        self.executor_id = executor_id
        self.result_dict = result_dict

    # make access dictionary-like, i.e. can do: result['pixels'], result['asset']
    def __getitem__(self, key):
        return self.get_result(key)

    def get_result(self, key):
        # different from BasicResult.get_result, it won't try to get
        # aggregate score
        return self.result_dict[key]

    def get_ordered_results(self):
        return sorted(self.result_dict.keys())


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import unittest
import os

import numpy as np

from vmaf.config import VmafConfig
from vmaf.core.asset import Asset
from vmaf.core.raw_extractor import AssetExtractor, DisYUVRawVideoExtractor

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class RawExtractorTest(unittest.TestCase):

    def test_run_asset_extractor(self):

        ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
        dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
        asset = Asset(dataset="test", content_id=0, asset_id=1,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=dis_path,
                      asset_dict={'width': 576, 'height': 324,
                                  'quality_width': 160, 'quality_height': 90})

        asset_original = Asset(dataset="test", content_id=0, asset_id=2,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=ref_path,
                      asset_dict={'width': 576, 'height': 324,
                                  'quality_width': 160, 'quality_height': 90})

        self.fextractor = AssetExtractor(
            [asset, asset_original], None, fifo_mode=True)

        self.fextractor.run()

        results = self.fextractor.results

        self.assertEqual(str(results[0]['asset']), 'test_0_1_src01_hrc00_576x324_576x324_vs_src01_hrc01_576x324_576x324_q_160x90')
        self.assertEqual(str(results[1]['asset']), 'test_0_2_src01_hrc00_576x324_576x324_vs_src01_hrc00_576x324_576x324_q_160x90')

        self.fextractor.run(parallelize=True)

        results = self.fextractor.results

        self.assertEqual(str(results[0]['asset']), 'test_0_1_src01_hrc00_576x324_576x324_vs_src01_hrc01_576x324_576x324_q_160x90')
        self.assertEqual(str(results[1]['asset']), 'test_0_2_src01_hrc00_576x324_576x324_vs_src01_hrc00_576x324_576x324_q_160x90')


class DisYUVRawVideoExtractorTest(unittest.TestCase):

    def setUp(self):
        self.h5py_filepath = VmafConfig.workdir_path('test.hdf5')

    def tearDown(self):
        if os.path.exists(self.h5py_filepath):
            os.remove(self.h5py_filepath)

    def test_run_dis_yuv_raw_video_extractor(self):
        ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
        dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
        asset = Asset(dataset="test", content_id=0, asset_id=1,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=dis_path,
                      asset_dict={'width': 576, 'height': 324})

        asset_original = Asset(dataset="test", content_id=0, asset_id=2,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=ref_path,
                      asset_dict={'width': 576, 'height': 324})

        h5py_file = DisYUVRawVideoExtractor.open_h5py_file(self.h5py_filepath)

        self.fextractor = DisYUVRawVideoExtractor(
            [asset, asset_original], None, fifo_mode=False,
            optional_dict={'channels': 'yu'},
            optional_dict2={'h5py_file': h5py_file}
        )

        self.fextractor.run()

        results = self.fextractor.results

        self.assertAlmostEqual(np.mean(results[0]['dis_y']), 61.332006579182384, places=4)
        self.assertAlmostEqual(np.mean(results[1]['dis_y']), 59.788567297525148, places=4)
        self.assertAlmostEqual(np.mean(results[0]['dis_u']), 115.23227407335962, places=4)
        self.assertAlmostEqual(np.mean(results[1]['dis_u']), 114.49701717535437, places=4)

        with self.assertRaises(KeyError):
            np.mean(results[0]['dis_v'])

        DisYUVRawVideoExtractor.close_h5py_file(h5py_file)

    def test_run_dis_yuv_raw_video_extractor_parallel(self):
        ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
        dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
        asset = Asset(dataset="test", content_id=0, asset_id=1,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=dis_path,
                      asset_dict={'width': 576, 'height': 324})

        asset_original = Asset(dataset="test", content_id=0, asset_id=2,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=ref_path,
                      asset_dict={'width': 576, 'height': 324})

        h5py_file = DisYUVRawVideoExtractor.open_h5py_file(self.h5py_filepath)

        self.fextractor = DisYUVRawVideoExtractor(
            [asset, asset_original], None, fifo_mode=False,
            optional_dict={'channels': 'yu'},
            optional_dict2={'h5py_file': h5py_file}
        )

        with self.assertRaises(AssertionError):
            self.fextractor.run(parallelize=True)

        DisYUVRawVideoExtractor.close_h5py_file(h5py_file)


class ParallelDisYRawVideoExtractorTest(unittest.TestCase):

    def setUp(self):
        self.h5py_filepath = VmafConfig.workdir_path('test.hdf5')

    def tearDown(self):
        if os.path.exists(self.h5py_filepath):
            os.remove(self.h5py_filepath)

    def test_run_parallel_dis_y_fextractor(self):
        ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
        dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
        asset = Asset(dataset="test", content_id=0, asset_id=1,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=dis_path,
                      asset_dict={'width': 576, 'height': 324})

        asset_original = Asset(dataset="test", content_id=0, asset_id=2,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=ref_path,
                      asset_dict={'width': 576, 'height': 324})

        h5py_file = DisYUVRawVideoExtractor.open_h5py_file(self.h5py_filepath)
        optional_dict2 = {'h5py_file': h5py_file}

        fextractor = DisYUVRawVideoExtractor(
            [asset, asset_original],
            None,
            fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict={'channels': 'yu'},
            optional_dict2=optional_dict2
        )
        self.fextractors = [fextractor]
        fextractor.run(parallelize=False)  # Can't run parallel: can't pickle FileID objects
        results = fextractor.results

        self.assertAlmostEqual(np.mean(results[0]['dis_y']), 61.332006579182384, places=4)
        self.assertAlmostEqual(np.mean(results[1]['dis_y']), 59.788567297525148, places=4)
        self.assertAlmostEqual(np.mean(results[0]['dis_u']), 115.23227407335962, places=4)
        self.assertAlmostEqual(np.mean(results[1]['dis_u']), 114.49701717535437, places=4)

        with self.assertRaises(KeyError):
            np.mean(results[0]['dis_v'])

        DisYUVRawVideoExtractor.close_h5py_file(h5py_file)


if __name__ == '__main__':
    unittest.main(verbosity=2)



__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

import os

import numpy as np

from vmaf.config import VmafConfig
from vmaf.core.executor import run_executors_in_parallel
from vmaf.core.raw_extractor import DisYUVRawVideoExtractor
from vmaf.core.nn_train_test_model import ToddNoiseClassifierTrainTestModel
from vmaf.routine import read_dataset
from vmaf.tools.misc import import_python_file


def main():
    # parameters
    num_train = 500
    num_test = 50
    n_epochs = 30
    seed = 0 # None

    # read input dataset
    dataset_path = VmafConfig.resource_path('dataset', 'BSDS500_noisy_dataset.py')
    dataset = import_python_file(dataset_path)
    assets = read_dataset(dataset)

    # shuffle assets
    np.random.seed(seed)
    np.random.shuffle(assets)
    assets = assets[:(num_train + num_test)]

    raw_video_h5py_filepath = VmafConfig.workdir_path('rawvideo.hdf5')
    raw_video_h5py_file = DisYUVRawVideoExtractor.open_h5py_file(raw_video_h5py_filepath)

    print('======================== Extract raw YUVs ==============================')

    _, raw_yuvs = run_executors_in_parallel(
        DisYUVRawVideoExtractor,
        assets,
        fifo_mode=True,
        delete_workdir=True,
        parallelize=False, # CAN ONLY USE SERIAL MODE FOR DisYRawVideoExtractor
        result_store=None,
        optional_dict=None,
        optional_dict2={'h5py_file': raw_video_h5py_file})

    patch_h5py_filepath = VmafConfig.workdir_path('patch.hdf5')
    patch_h5py_file = ToddNoiseClassifierTrainTestModel.open_h5py_file(patch_h5py_filepath)
    model = ToddNoiseClassifierTrainTestModel(
        param_dict={
            'seed': seed,
            'n_epochs': n_epochs,
        },
        logger=None,
        optional_dict2={ # for options that won't impact the result
            # 'checkpoints_dir': VmafConfig.workspace_path('checkpoints_dir'),
            'h5py_file': patch_h5py_file,
        })

    print('============================ Train model ===============================')
    xys = ToddNoiseClassifierTrainTestModel.get_xys_from_results(raw_yuvs[:num_train])
    model.train(xys)

    print('=========================== Evaluate model =============================')
    xs = ToddNoiseClassifierTrainTestModel.get_xs_from_results(raw_yuvs[num_train:])
    ys = ToddNoiseClassifierTrainTestModel.get_ys_from_results(raw_yuvs[num_train:])
    result = model.evaluate(xs, ys)

    print("")
    print("f1 test %g, errorrate test %g" % (result['f1'], result['errorrate']))

    # tear down
    DisYUVRawVideoExtractor.close_h5py_file(raw_video_h5py_file)
    ToddNoiseClassifierTrainTestModel.close_h5py_file(patch_h5py_file)
    os.remove(raw_video_h5py_filepath)
    os.remove(patch_h5py_filepath)

    print('Done.')


if __name__ == "__main__":
    main()



dataset_name = 'example'

yuv_fmt = 'yuv420p'
width = 576
height = 324
quality_width = 576
quality_height = 324
duration_sec = 5.0

from vmaf.config import VmafConfig

ref_videos = [
    {'content_id': 0, 'path': VmafConfig.test_resource_path('yuv', 'src01_hrc00_576x324.yuv')},
]

dis_videos = [
    {'content_id': 0, 'asset_id': 1, 'dmos': 50, 'path': VmafConfig.test_resource_path('yuv', 'src01_hrc01_576x324.yuv'), 'crop_cmd': '288:162:144:81'},
    {'content_id': 0, 'asset_id': 2, 'dmos': 49, 'path': VmafConfig.test_resource_path('yuv', 'src01_hrc01_576x324.yuv'), 'pad_cmd': 'iw+100:ih+100:50:50'},
    {'content_id': 0, 'asset_id': 3, 'dmos': 48, 'path': VmafConfig.test_resource_path('yuv', 'src01_hrc01_576x324.yuv'), 'crop_cmd': '288:162:144:81', 'pad_cmd': 'iw+288:ih+162:144:81',},
]



import numpy as np

import sklearn.metrics
from sklearn.linear_model import Ridge

from vmaf import plt
from vmaf.tools.misc import get_file_name_without_extension
from vmaf.tools.reader import YuvReader

# Copyright (c) 2016, Marco Tulio Correia Ribeiro
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


class LocalExplainer(object):
    """Explains a TrainTestModel on a local data point.
    Adapted from:
    Lime: Explaining the predictions of any machine learning classifier
    https://github.com/marcotcr/lime"""

    def __init__(self,
                 neighbor_std=1.0,
                 neighbor_samples=5000,
                 distance_metric='euclidean',
                 kernel_width=3,
                 model_regressor=None,
                 ):
        """Init function.

        :param neighbor_std: standard deviation of neighborhood sampled
        :param neighbor_samples: number of samples of neighborhood
        :param distance_metric: distance metric used
        :param kernel_width: width for kernel function
        :param model_regressor: regressor to train local linear model. If None, use Ridge
        """
        self.neighbor_std = neighbor_std
        self.neighbor_samples = neighbor_samples
        self.distance_metric = distance_metric
        self.kernel_fn = lambda d: np.sqrt(np.exp(-(d**2) / kernel_width ** 2))
        self.model_regressor = Ridge(alpha=1, fit_intercept=True) if model_regressor is None else model_regressor

    def _assert_model(self, train_test_model):

        train_test_model._assert_trained()

        assert hasattr(train_test_model, '_to_tabular_xs'), \
            'train_test_model must have a method _to_tabular_xs().'

        assert train_test_model.norm_type != 'none', \
            'If train_test_model has not been normalized, ' \
            'the sampled neighborhood may not be of the right shape.'

    def explain(self, train_test_model, xs):
        """Explain data points.

        :param train_test_model: a trained TrainTestModel object
        :param xs: same xs as in TrainTestModel.predict(xs)
        :return: exps: explanations, where exps['feature_weights'] has the
        feature weights (in num_sample x num_feature 2D array)
        """

        self._assert_model(train_test_model)

        feature_names = train_test_model.feature_names
        for name in feature_names:
            assert name in xs

        xs_2d = train_test_model._to_tabular_xs(feature_names, xs)

        xs_2d_unnormalized = xs_2d.copy()

        # normalize xs
        xs_2d = train_test_model.normalize_xs(xs_2d)

        # for each row of xs, repeat feature of a unit (e.g. frame),
        # generate a new 2d_array by sampling its neighborhood
        n_sample, n_feature = xs_2d.shape
        feature_weights = np.zeros([n_sample, n_feature])
        for i_sample in range(n_sample):

            # generate neighborhood samples
            x_row = xs_2d[i_sample, :]
            xs_2d_neighbor = np.random.randn(self.neighbor_samples, n_feature) * self.neighbor_std
            xs_2d_neighbor += np.tile(x_row, (self.neighbor_samples, 1))

            # add center to first row
            xs_2d_neighbor = np.vstack([x_row, xs_2d_neighbor])

            # calculate distance to center
            distances = sklearn.metrics.pairwise_distances(
                xs_2d_neighbor,
                xs_2d_neighbor[0].reshape(1, -1),
                metric=self.distance_metric
            ).ravel()
            sample_weight = self.kernel_fn(distances)

            model = train_test_model.model
            if isinstance(model, list):
                model = model[0]  # HACKY, TODO: fix it

            # predict
            ys_label_pred_neighbor = train_test_model._predict(model, xs_2d_neighbor)

            # take xs_2d_neighbor and ys_label_pred_neighbor, train a linear
            # model
            self.model_regressor.fit(xs_2d_neighbor,
                                     ys_label_pred_neighbor,
                                     sample_weight=sample_weight)
            feature_weight = self.model_regressor.coef_.copy()
            feature_weights[i_sample, :] = feature_weight

        exps = {
            'feature_weights': feature_weights,
            'features': xs_2d_unnormalized,
            'features_normalized': xs_2d,
            'feature_names': feature_names
        }

        return exps

    @staticmethod
    def assert_explanations(exps, assets=None, ys=None, ys_pred=None):
        N = exps['feature_weights'].shape[0]
        assert N == exps['features_normalized'].shape[0]
        if assets is not None:
            assert N == len(assets)
        if ys is not None:
            assert N == len(ys['label'])
        if ys_pred is not None:
            assert N == len(ys_pred)
        return N

    @classmethod
    def print_explanations(cls, exps, assets=None, ys=None, ys_pred=None):

        # asserts
        N = cls.assert_explanations(exps, assets, ys, ys_pred)

        print("Features: {}".format(exps['feature_names']))

        for n in range(N):
            weights = exps['feature_weights'][n]
            features = exps['features_normalized'][n]

            asset = assets[n] if assets is not None else None
            y = ys['label'][n] if ys is not None else None
            y_pred = ys_pred[n] if ys_pred is not None else None

            print("{ref}".format(
                ref=get_file_name_without_extension(asset.ref_path) if
                asset is not None else "Asset {}".format(n)))
            if asset is not None:
                print("\tDistorted: {dis}".format(
                    dis=get_file_name_without_extension(asset.dis_path)))
            if y is not None:
                print("\tground truth: {y:.3f}".format(y=y))
            if y_pred is not None:
                print("\tpredicted: {y_pred:.3f}".format(y_pred=y_pred))
            print("\tfeature value: {}".format(features))
            print("\tfeature weight: {}".format(weights))

    @classmethod
    def plot_explanations(cls, exps, assets=None, ys=None, ys_pred=None):

        # asserts
        N = cls.assert_explanations(exps, assets, ys, ys_pred)

        figs = []
        for n in range(N):
            weights = exps['feature_weights'][n]
            features = exps['features'][n]
            normalized = exps['features_normalized'][n]

            asset = assets[n] if assets is not None else None
            y = ys['label'][n] if ys is not None else None
            y_pred = ys_pred[n] if ys_pred is not None else None

            img = None
            if asset is not None:
                w, h = asset.dis_width_height
                with YuvReader(filepath=asset.dis_path, width=w, height=h,
                               yuv_type=asset.dis_yuv_type) as yuv_reader:
                    for yuv in yuv_reader:
                        img, _, _ = yuv
                        img = img.astype(np.double)
                        break
                assert img is not None

            title = ""
            if asset is not None:
                title += "{}\n".format(get_file_name_without_extension(asset.ref_path))
            if y is not None:
                title += "ground truth: {:.3f}\n".format(y)
            if y_pred is not None:
                title += "predicted: {:.3f}\n".format(y_pred)
            if title != "" and title[-1] == '\n':
                title = title[:-1]

            assert len(weights) == len(features)
            M = len(weights)

            fig = plt.figure()

            ax_top = plt.subplot(2, 1, 1)
            ax_left = plt.subplot(2, 3, 4)
            ax_mid = plt.subplot(2, 3, 5, sharey=ax_left)
            ax_right = plt.subplot(2, 3, 6, sharey=ax_left)

            if img is not None:
                ax_top.imshow(img, cmap='Greys_r')
            ax_top.get_xaxis().set_visible(False)
            ax_top.get_yaxis().set_visible(False)
            ax_top.set_title(title)

            pos = np.arange(M) + 0.1
            ax_left.barh(pos, features, color='b', label='feature')
            ax_left.set_xticks(np.arange(0, 1.1, 0.2))
            ax_left.set_yticks(pos + 0.35)
            ax_left.set_yticklabels(exps['feature_names'])
            ax_left.set_title('feature')

            ax_mid.barh(pos, normalized, color='g', label='fnormal')
            ax_mid.get_yaxis().set_visible(False)
            ax_mid.set_title('fnormal')

            ax_right.barh(pos, weights, color='r', label='weight')
            ax_right.get_yaxis().set_visible(False)
            ax_right.set_title('weight')

            plt.tight_layout()

            figs.append(fig)

        return figs

    @classmethod
    def select_from_exps(cls, exps, indexs):
        # asserts
        N = cls.assert_explanations(exps)
        for index in indexs:
            assert index < N
        exps2 = {
            'feature_weights': exps['feature_weights'][indexs, :],
            'features': exps['features'][indexs, :],
            'features_normalized': exps['features_normalized'][indexs, :],
            'feature_names': exps['feature_names']
        }
        return exps2



from abc import ABC

from vmaf.core.cambi_feature_extractor import CambiFeatureExtractor, CambiFullReferenceFeatureExtractor
from vmaf.core.quality_runner import QualityRunnerFromFeatureExtractor
from vmaf.tools.decorator import override


class CambiQualityRunner(QualityRunnerFromFeatureExtractor, ABC):

    TYPE = 'Cambi'
    VERSION = 'F' + CambiFeatureExtractor.VERSION

    @override(QualityRunnerFromFeatureExtractor)
    def _get_feature_extractor_class(self):
        return CambiFeatureExtractor

    @override(QualityRunnerFromFeatureExtractor)
    def _get_feature_key_for_score(self):
        return 'cambi'


class CambiFullReferenceQualityRunner(QualityRunnerFromFeatureExtractor, ABC):

    TYPE = 'Cambi_FR'
    VERSION = 'F' + CambiFeatureExtractor.VERSION

    @override(QualityRunnerFromFeatureExtractor)
    def _get_feature_extractor_class(self):
        return CambiFullReferenceFeatureExtractor

    @override(QualityRunnerFromFeatureExtractor)
    def _get_feature_key_for_score(self):
        return 'cambi_full_reference'



import re
from typing import List, Optional

from vmaf.tools.misc import MyTestCase


def get_tidy_mock_call_args_list(mockProcessRunner_run) -> List[str]:
    l = list()
    for i in range(len(mockProcessRunner_run.call_args_list)):
        e = mockProcessRunner_run.call_args_list[i][0][0]
        if isinstance(e, str):
            l.append(e)
        else:
            l.append(' '.join(e))
    return l


def replace_uuid(command_line: str) -> str:
    """
    Replace UUIDs in a command line with pattern [UUID]

    >>> replace_uuid('/tmp/72b3a7af-204c-4455-afe5-be2d536f2fdd/dv_el_out_1.h265')
    '/tmp/[UUID]/dv_el_out_1.h265'
    """
    uuid_pattern = r'\b[a-f\d]{8}(?:-[a-f\d]{4}){3}-[a-f\d]{12}\b'
    return re.sub(uuid_pattern, '[UUID]', command_line)


def replace_root(command_line: str, root: str) -> str:
    """
    Replace root directory specified in input with pattern [ROOT]

    >>> replace_root('/opt/project/vmaf/libvmaf/build/tools/vmaf', root='/opt/project')
    '[ROOT]/vmaf/libvmaf/build/tools/vmaf'
    >>> replace_root('/tmp/72b3a7af-204c-4455-afe5-be2d536f2fdd', root='/opt/project')
    '/tmp/72b3a7af-204c-4455-afe5-be2d536f2fdd'
    """
    return re.sub(root, '[ROOT]', command_line)


def remove_redundant_whitespace(command_line: str) -> str:
    """
    Replaces multiple whitespace between words with a single one, and removes redundant whitespace at the start and end
    >>> remove_redundant_whitespace('  a b  c d   e     f ')
    'a b c d e f'
    >>> remove_redundant_whitespace('cat /opt/project/vmaf/workspace/workdir/9e693ccc-7706-49c5-8c8e-40f5242e81a6/dis_test_0_0_seeking_10_288_375_notyuv_lanczos_accurate_rnd_10to14_prece_FFmpegDecoder_postunsharpunsharp_q_480x360_PostDecode_tmp/pixfmt/* >  /opt/project/vmaf/workspace/workdir/9e693ccc-7706-49c5-8c8e-40f5242e81a6/dis_test_0_0_seeking_10_288_375_notyuv_lanczos_accurate_rnd_10to14_prece_FFmpegDecoder_postunsharpunsharp_q_480x360_PostPreresamplingFilter0 ')
    'cat /opt/project/vmaf/workspace/workdir/9e693ccc-7706-49c5-8c8e-40f5242e81a6/dis_test_0_0_seeking_10_288_375_notyuv_lanczos_accurate_rnd_10to14_prece_FFmpegDecoder_postunsharpunsharp_q_480x360_PostDecode_tmp/pixfmt/* > /opt/project/vmaf/workspace/workdir/9e693ccc-7706-49c5-8c8e-40f5242e81a6/dis_test_0_0_seeking_10_288_375_notyuv_lanczos_accurate_rnd_10to14_prece_FFmpegDecoder_postunsharpunsharp_q_480x360_PostPreresamplingFilter0'
    """
    return " ".join(command_line.split())


def remove_option(command_line: str, option: str) -> str:
    """
    Removes a whitespace-separated option that is prefixed by two dashes, e.g., --option_name.
    >>> remove_option('vmaf --reference REFERENCE --model MODEL', 'model')
    'vmaf --reference REFERENCE'
    >>> remove_option('vmaf --model MODEL --reference REFERENCE', 'model')
    'vmaf --reference REFERENCE'
    >>> remove_option(remove_option('vmaf --model MODEL --dist DIST --reference REFERENCE', 'model'), 'reference')
    'vmaf --dist DIST'
    >>> remove_option(remove_option('vmaf --model MODEL --dist DIST --reference REFERENCE', 'reference'), 'model')
    'vmaf --dist DIST'
    >>> remove_option('vmaf --reference REFERENCE', 'model')
    'vmaf --reference REFERENCE'
    >>> remove_option('a --model M b', 'model')
    'a b'
    >>> remove_option('a --model KLM b', 'model')
    'a b'
    >>> remove_option('abc --model K def', 'model')
    'abc def'
    >>> remove_option('abc --model KLM def', 'model')
    'abc def'
    >>> remove_option('abc --model KLM d', 'model')
    'abc d'
    >>> remove_option('a --model KLM def', 'model')
    'a def'
    >>> remove_option('a --model M b c --model M d', 'model')
    'a b c d'
    >>> remove_option('a --model M', 'model')
    'a'
    >>> remove_option('a --model M ', 'model')
    'a '
    >>> remove_option(' --model M ', 'model')
    ' '
    >>> remove_option('--model M ', 'model')
    ' '
    >>> remove_option('--model2 M ', 'model')
    '--model2 M '
    >>> remove_option('-- model M ', 'model')
    '-- model M '
    """
    if command_line.startswith('--{option}'.format(option=option)):
        return re.sub(r'--{option} [^\s]*'.format(option=option), '', command_line)
    else:
        return re.sub(r' --{option} [^\s]*'.format(option=option), '', command_line)


def remove_elements_containing_substring(command_line: str, sub_str: str) -> str:
    """
    Removes strings from the command line that contain a specific substring
    >>> remove_elements_containing_substring('cat /opt/project/vmaf/workspace/workdir/9e693ccc-7706-49c5-8c8e-40f5242e81a6/dis_test_0_0_seeking_10_288_375_notyuv_lanczos_accurate_rnd_10to14_prece_FFmpegDecoder_postunsharpunsharp_q_480x360_PostDecode_tmp/pixfmt/* >  /opt/project/vmaf/workspace/workdir/9e693ccc-7706-49c5-8c8e-40f5242e81a6/dis_test_0_0_seeking_10_288_375_notyuv_lanczos_accurate_rnd_10to14_prece_FFmpegDecoder_postunsharpunsharp_q_480x360_PostPreresamplingFilter0', 'workspace/workdir')
    'cat >'
    """
    assert isinstance(sub_str, str)
    return " ".join([x for x in command_line.split() if sub_str not in x])


def assert_equivalent_commands(self, cmds: List[str], cmds_expected: List[str], root: str, root_expected: str, do_replace_uuid: bool = True,
                               options_to_remove: Optional[List[str]] = None, substrings_to_remove: Optional[List[str]] = None):
    """
    >>> self = MyTestCase()
    >>> self.setUp()
    >>> assert_equivalent_commands(self, cmds=["/opt/project/vmaf/libvmaf/build/tools/vmaf /tmp/72b3a7af-204c-4455-afe5-be2d536f2fdd/dv_el_out_1.h265"], cmds_expected=["/opt/project/vmaf/libvmaf/build/tools/vmaf /tmp/82b3a7af-304c-5455-afe5-be2d536f2fdd/dv_el_out_1.h265"], root="/opt/project", root_expected="/opt/project")
    >>> self.tearDown()
    >>> self2 = MyTestCase()
    >>> self2.setUp()
    >>> assert_equivalent_commands(self2, cmds=["/opt/project/vmaf/libvmaf/build/tools/vmaf /tmp/72b3a7af-204c-4455-afe5-be2d536f2fdd/dv_el_out_1.h265"], cmds_expected=["/opt/project/vmaf/libvmaf/build/tools/vmaf /tmp/82b3a7af-304c-5455-afe5-be2d536f2fdd/dv_el_out_1.h265"], root="/opt/project", root_expected="/opt/project", do_replace_uuid=False)
    >>> with self.assertRaises(AssertionError): self2.tearDown()
    >>> self3 = MyTestCase()
    >>> self3.setUp()
    >>> assert_equivalent_commands(self3, cmds=["/opt/project/vmaf --reference ref.h265 --distorted dist.h265"], cmds_expected=["/opt/project/vmaf --reference ref.h265 --distorted dist.h266"], root="/opt/project", root_expected="/opt/project", options_to_remove=["distorted"])
    >>> self3.tearDown()
    >>> self4 = MyTestCase()
    >>> self4.setUp()
    >>> assert_equivalent_commands(self4, cmds=["/opt/project/vmaf --reference ref.h265 --distorted dist.h265 --output output.xml"], cmds_expected=["/opt/project/vmaf --reference ref.h266 --distorted dist.h266 --output output.xml"], root="/opt/project", root_expected="/opt/project", options_to_remove=["distorted", "reference"])
    >>> self4.tearDown()
    >>> self5 = MyTestCase()
    >>> self5.setUp()
    >>> assert_equivalent_commands(self5, cmds=["/opt/project/vmaf --reference /opt/project/vmaf/workspace/workdir/ref.h265 --distorted /opt/project/vmaf/workspace/workdir/dist.h265 --output output.xml"], cmds_expected=["/opt/project/vmaf --reference /opt/project/vmaf/workspace/workdir/ref.h266 --distorted /opt/project/vmaf/workspace/workdir/dist.h266 --output output2.xml"], root="/opt/project", root_expected="/opt/project", substrings_to_remove=["workspace/workdir", "output"])
    >>> self5.tearDown()
    """

    if options_to_remove is None:
        options_to_remove = []
    if substrings_to_remove is None:
        substrings_to_remove = []
    assert len(cmds) == len(cmds_expected), f"length of cmds and cmds_expected are not equal: {len(cmds)} vs. {len(cmds_expected)}"
    for cmd, cmd_expected in zip(cmds, cmds_expected):

        if do_replace_uuid is True:
            cmd1 = replace_uuid(cmd)
        else:
            cmd1 = cmd
        cmd2 = replace_root(cmd1, root)
        cmd3 = remove_redundant_whitespace(cmd2)
        for option_to_remove in options_to_remove:
            cmd3 = remove_option(cmd3, option_to_remove)
        for sbstr_to_remove in substrings_to_remove:
            cmd3 = remove_elements_containing_substring(cmd3, sbstr_to_remove)

        cmd_expected1 = replace_uuid(cmd_expected)
        cmd_expected2 = replace_root(cmd_expected1, root_expected)
        cmd_expected3 = remove_redundant_whitespace(cmd_expected2)
        for option_to_remove in options_to_remove:
            cmd_expected3 = remove_option(cmd_expected3, option_to_remove)
        for sbstr_to_remove in substrings_to_remove:
            cmd_expected3 = remove_elements_containing_substring(cmd_expected3, sbstr_to_remove)

        self.assertEqual(cmd3, cmd_expected3, msg=f"cmd and cmd_expected are not matched:\ncmd: {cmd}\ncmd:expected: "
                                                  f"{cmd_expected}\nprocessed cmd: {cmd3}\nprocessed cmd:expected: "
                                                  f"{cmd_expected3}")



__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

import unittest

from vmaf.config import VmafConfig
from vmaf.core.asset import Asset, NorefAsset


def set_default_flat_1920_1080_videos_for_testing():
    ref_path = VmafConfig.test_resource_path("yuv", "flat_1920_1080_0.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "flat_1920_1080_10.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_path,
                  dis_path=dis_path,
                  asset_dict={'width': 1920, 'height': 1080})

    asset_original = Asset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           ref_path=ref_path,
                           dis_path=ref_path,
                           asset_dict={'width': 1920, 'height': 1080})

    return ref_path, dis_path, asset, asset_original


def set_default_576_324_videos_for_testing():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_path,
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324})

    asset_original = Asset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           ref_path=ref_path,
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324})

    return ref_path, dis_path, asset, asset_original

def set_default_576_324_videos_for_testing_workfile_yuv_10b():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_path,
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324, 'workfile_yuv_type': 'yuv420p10le'})

    asset_original = Asset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           ref_path=ref_path,
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324, 'workfile_yuv_type': 'yuv420p10le'})

    return ref_path, dis_path, asset, asset_original

def set_default_576_324_videos_for_testing_scaled():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_path,
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324,
                              'dis_enc_width': 480, 'dis_enc_height': 270})

    asset_original = Asset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           ref_path=ref_path,
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324,
                                       'dis_enc_width': 480, 'dis_enc_height': 270})

    return ref_path, dis_path, asset, asset_original


def set_default_576_324_10bit_videos_for_testing():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv422p10le.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv422p10le.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_path,
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324,
                              'yuv_type': 'yuv422p10le'})

    asset_original = Asset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           ref_path=ref_path,
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324,
                                       'yuv_type': 'yuv422p10le'})

    return ref_path, dis_path, asset, asset_original


def set_default_576_324_10bit_videos_for_testing_b():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv420p10le.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv420p10le.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_path,
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324,
                              'yuv_type': 'yuv420p10le'})

    asset_original = Asset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           ref_path=ref_path,
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324,
                                       'yuv_type': 'yuv420p10le'})

    return ref_path, dis_path, asset, asset_original


def set_default_576_324_12bit_videos_for_testing():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv420p12le.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv420p12le.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_path,
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324,
                              'yuv_type': 'yuv420p12le'})

    asset_original = Asset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           ref_path=ref_path,
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324,
                                       'yuv_type': 'yuv420p12le'})

    return ref_path, dis_path, asset, asset_original


def set_default_576_324_16bit_videos_for_testing():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv420p16le.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv420p16le.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=ref_path,
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324,
                              'yuv_type': 'yuv420p16le'})

    asset_original = Asset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           ref_path=ref_path,
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324,
                                       'yuv_type': 'yuv420p16le'})

    return ref_path, dis_path, asset, asset_original


def set_default_576_324_noref_videos_for_testing():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
    asset = NorefAsset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324})

    asset_original = NorefAsset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324})

    return ref_path, dis_path, asset, asset_original

def set_default_576_324_noref_videos_for_testing_workfile_yuv_10b():
    ref_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv")
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc01_576x324.yuv")
    asset = NorefAsset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324, 'workfile_yuv_type': 'yuv420p10le'})

    asset_original = NorefAsset(dataset="test", content_id=0, asset_id=1,
                           workdir_root=VmafConfig.workdir_path(),
                           dis_path=ref_path,
                           asset_dict={'width': 576, 'height': 324, 'workfile_yuv_type': 'yuv420p10le'})

    return ref_path, dis_path, asset, asset_original

def set_default_cambi_video_for_testing():
    dis_path = VmafConfig.test_resource_path("yuv", "blue_sky_360p_60f.yuv")
    asset = NorefAsset(dataset="test", content_id=0, asset_id=0,
                       workdir_root=VmafConfig.workdir_path(),
                       dis_path=dis_path,
                       asset_dict={'width': 640, 'height': 360, 'yuv_type': 'yuv420p'})

    return dis_path, asset


def set_default_cambi_video_for_testing_b():
    dis_path = VmafConfig.test_resource_path("yuv", "KristenAndSara_1280x720_8bit_processed.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=dis_path,
                  dis_path=dis_path,
                  asset_dict={'width': 1280, 'height': 720,
                              'dis_enc_width': 960, 'dis_enc_height': 540})

    return dis_path, dis_path, asset, asset

def set_default_cambi_video_for_testing_10b():
    dis_path = VmafConfig.test_resource_path("yuv", "src01_hrc00_576x324.yuv420p10le.yuv")
    asset = Asset(dataset="test", content_id=0, asset_id=0,
                  workdir_root=VmafConfig.workdir_path(),
                  ref_path=dis_path,
                  dis_path=dis_path,
                  asset_dict={'width': 576, 'height': 324,
                              'yuv_type': 'yuv420p10le'})

    return dis_path, dis_path, asset, asset



from vmaf.core.local_explainer import LocalExplainer
from vmaf.core.quality_runner import VmafQualityRunner
from vmaf.core.result import Result
from vmaf.tools.decorator import override

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class VmafQualityRunnerWithLocalExplainer(VmafQualityRunner):
    """Same as VmafQualityRunner, except that it outputs additional LocalExplainer
    results. It shares the type 'VMAF' with VmafQualityRunner, since the numeric
    result generated is essentially the same. But on the other hand, it doesn't
    want to be searchable by type 'VMAF', so it is put in a different module file.
    """

    TYPE = 'VMAF_LE'
    VERSION = '{}-le1'.format(VmafQualityRunner.VERSION)

    @classmethod
    def get_explanations_key(cls):
        return cls.get_scores_key() + '_exps'

    @override(VmafQualityRunner)
    def _run_on_asset(self, asset):
        # Override VmafQualityRunner._run_on_asset(self, asset), by adding
        # additional local explanation info.
        vmaf_fassembler = self._get_vmaf_feature_assembler_instance(asset)
        vmaf_fassembler.run()
        feature_result = vmaf_fassembler.results[0]
        model = self._load_model(asset)
        xs = model.get_per_unit_xs_from_a_result(feature_result)
        ys_pred = self.predict_with_model(model, xs)['ys_pred']

        if self.optional_dict2 is not None and \
           'explainer' in self.optional_dict2:
            explainer = self.optional_dict2['explainer']
        else:
            explainer = LocalExplainer()

        exps = explainer.explain(model, xs)
        result_dict = {}
        result_dict.update(feature_result.result_dict)  # add feature result
        result_dict[self.get_scores_key()] = ys_pred  # add quality score
        result_dict[self.get_explanations_key()] = exps  # add local explanations
        return Result(asset, self.executor_id, result_dict)

    @classmethod
    def show_local_explanations(cls, results, indexes=None):
        """Plot local explanations of results

        :param results:
        :param indexes: a list of frame indices, or None. If None, will take the
        second frame.
        :return: figures of local explanation plots
        """

        # assert results are indeed generated by class
        for result in results:
            assert cls.get_explanations_key() in result.result_dict

        N = len(results)

        if indexes is None:
            indexes = [1]  # default: second frame

        figss = []
        for n in range(N):

            exps = results[n][cls.get_explanations_key()]
            asset = results[n].asset
            exps2 = LocalExplainer.select_from_exps(exps, indexes)

            ys_pred = results[n][cls.get_scores_key()][indexes]

            N2 = LocalExplainer.assert_explanations(exps2)
            assets2 = [asset for _ in range(N2)]

            # LocalExplainer.print_explanations(exps2, assets=assets2, ys=None, ys_pred=ys_pred)
            figs = LocalExplainer.plot_explanations(exps2, assets=assets2, ys=None, ys_pred=ys_pred)
            figss.append(figs)

        return figss



import os
import unittest

import numpy as np

from vmaf.config import VmafConfig
from vmaf.core.train_test_model import TrainTestModel, \
    LibsvmNusvrTrainTestModel, SklearnRandomForestTrainTestModel, \
    MomentRandomForestTrainTestModel, SklearnExtraTreesTrainTestModel, \
    SklearnLinearRegressionTrainTestModel, Logistic5PLRegressionTrainTestModel
from vmaf.core.noref_feature_extractor import MomentNorefFeatureExtractor
from vmaf.routine import read_dataset
from vmaf.tools.misc import import_python_file
from vmaf.core.raw_extractor import DisYUVRawVideoExtractor

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class TrainTestModelTest(unittest.TestCase):

    def setUp(self):

        train_dataset_path = VmafConfig.test_resource_path('test_image_dataset_diffdim2.py')
        train_dataset = import_python_file(train_dataset_path)
        train_assets = read_dataset(train_dataset)

        runner = MomentNorefFeatureExtractor(
            train_assets,
            None,
            fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict=None,
            optional_dict2=None,
        )
        runner.run(parallelize=True)
        self.features = runner.results

        self.model_filename = VmafConfig.workspace_path("model", "test_save_load.pkl")

    def tearDown(self):
        if hasattr(self, 'model'):
            self.model.delete(self.model_filename)

    def test_get_xs_ys(self):
        xs = TrainTestModel.get_xs_from_results(self.features, [0, 1, 2])

        self.assertEqual(len(xs['Moment_noref_feature_1st_score']), 3)
        self.assertAlmostEqual(np.mean(xs['Moment_noref_feature_1st_score']), 128.26146851380497, places=4)
        self.assertEqual(len(xs['Moment_noref_feature_var_score']), 3)
        self.assertAlmostEqual(np.mean(xs['Moment_noref_feature_var_score']), 1569.2395085695462, places=4)

        xs = TrainTestModel.get_xs_from_results(self.features)
        self.assertEqual(len(xs['Moment_noref_feature_1st_score']), 9)
        self.assertAlmostEqual(np.mean(xs['Moment_noref_feature_1st_score']), 111.59099599173773, places=4)
        self.assertEqual(len(xs['Moment_noref_feature_var_score']), 9)
        self.assertAlmostEqual(np.mean(xs['Moment_noref_feature_var_score']), 1806.8620377229011, places=4)

        ys = TrainTestModel.get_ys_from_results(self.features, [0, 1, 2])
        expected_ys = {'label': np.array([2.5, 3.9, 5.0]),
                       'content_id': np.array([0, 1, 2])}
        self.assertTrue(all(ys['label'] == expected_ys['label']))
        self.assertTrue(all(ys['content_id'] == expected_ys['content_id']))

    def test_train_save_load_predict(self):

        xs = SklearnRandomForestTrainTestModel.get_xs_from_results(self.features)
        ys = SklearnRandomForestTrainTestModel.get_ys_from_results(self.features)
        xys = SklearnRandomForestTrainTestModel.get_xys_from_results(self.features)

        self.model = SklearnRandomForestTrainTestModel({'norm_type': 'normalize',
                                                        'n_estimators': 10,
                                                        'random_state': 0}, None)
        self.model.train(xys)

        self.model.to_file(self.model_filename)
        self.assertTrue(os.path.exists(self.model_filename))

        loaded_model = SklearnRandomForestTrainTestModel.from_file(self.model_filename, None)

        result = loaded_model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.17634739353518517, places=4)

    def test_train_save_load_predict_libsvmnusvr(self):

        xs = LibsvmNusvrTrainTestModel.get_xs_from_results(self.features)
        ys = LibsvmNusvrTrainTestModel.get_ys_from_results(self.features)
        xys = LibsvmNusvrTrainTestModel.get_xys_from_results(self.features)

        self.model = LibsvmNusvrTrainTestModel({'norm_type': 'normalize'}, None)
        self.model.train(xys)

        self.model.to_file(self.model_filename)
        self.assertTrue(os.path.exists(self.model_filename))
        self.assertTrue(os.path.exists(self.model_filename + '.model'))

        loaded_model = LibsvmNusvrTrainTestModel.from_file(self.model_filename, None)

        result = self.model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.62263086620058783, places=4)

        # loaded model generates slight numerical difference
        result = loaded_model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.62263139871631323, places=4)

    def test_train_predict_libsvmnusvr(self):

        # libsvmnusvr is bit exact to nusvr

        xs = LibsvmNusvrTrainTestModel.get_xs_from_results(self.features)
        ys = LibsvmNusvrTrainTestModel.get_ys_from_results(self.features)
        xys = LibsvmNusvrTrainTestModel.get_xys_from_results(self.features)

        model = LibsvmNusvrTrainTestModel(
            {'norm_type': 'normalize'}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.62263086620058783, places=4)

        model = LibsvmNusvrTrainTestModel(
            {'norm_type': 'clip_0to1'}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.84243141087114626, places=4)

        model = LibsvmNusvrTrainTestModel(
            {'norm_type': 'custom_clip_0to1',
             'custom_clip_0to1_map': {
                'Moment_noref_feature_1st_score': [0.0, 100.0],
              },
             }, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.84639162766546994, places=4)

        model = LibsvmNusvrTrainTestModel(
            {'norm_type': 'clip_minus1to1'}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.8314352752340991, places=4)

        model = LibsvmNusvrTrainTestModel(
            {'norm_type': 'none'}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.23294283650716496, places=4)

    def test_train_across_test_splits_ci_libsvmnusvr(self):

        xs = LibsvmNusvrTrainTestModel.get_xs_from_results(self.features)
        ys = LibsvmNusvrTrainTestModel.get_ys_from_results(self.features)
        xys = LibsvmNusvrTrainTestModel.get_xys_from_results(self.features)

        self.model = LibsvmNusvrTrainTestModel({'norm_type': 'normalize'}, None)
        self.model.train(xys)

        ys_label_pred = self.model.predict(xs)['ys_label_pred']
        ys_label = ys['label']

        n_splits_test_indices = 3

        stats = self.model.get_stats(ys_label, ys_label_pred,
                                     split_test_indices_for_perf_ci=True,
                                     n_splits_test_indices=n_splits_test_indices)

        # check that the performance metric distributions have been passed out
        assert 'SRCC_across_test_splits_distribution' in stats, 'SRCC across_test_splits distribution non-existing.'
        assert 'PCC_across_test_splits_distribution' in stats, 'PCC across_test_splits distribution non-existing.'
        assert 'RMSE_across_test_splits_distribution' in stats, 'RMSE across_test_splits distribution non-existing.'

        # check that the length of the perf metrc lists is equal to n_splits_test_indices
        assert len(stats['SRCC_across_test_splits_distribution']) == n_splits_test_indices, \
            'SRCC list is not equal to the number of splits specified.'
        assert len(stats['PCC_across_test_splits_distribution']) == n_splits_test_indices, \
            'PCC list is not equal to the number of splits specified.'
        assert len(stats['RMSE_across_test_splits_distribution']) == n_splits_test_indices, \
            'RMSE list is not equal to the number of splits specified.'

        self.assertAlmostEqual(stats['SRCC_across_test_splits_distribution'][0], 0.391304347826087, places=4)
        self.assertAlmostEqual(stats['SRCC_across_test_splits_distribution'][1], 0.8983050847457626, places=4)
        self.assertAlmostEqual(stats['SRCC_across_test_splits_distribution'][2], 0.9478260869565218, places=4)

        self.assertAlmostEqual(stats['PCC_across_test_splits_distribution'][0], 0.554154433495891, places=4)
        self.assertAlmostEqual(stats['PCC_across_test_splits_distribution'][1], 0.817203617522247, places=4)
        self.assertAlmostEqual(stats['PCC_across_test_splits_distribution'][2], 0.5338890441054945, places=4)

        self.assertAlmostEqual(stats['RMSE_across_test_splits_distribution'][0], 0.6943573719335835, places=4)
        self.assertAlmostEqual(stats['RMSE_across_test_splits_distribution'][1], 0.5658403884750773, places=4)
        self.assertAlmostEqual(stats['RMSE_across_test_splits_distribution'][2], 0.5997800884581888, places=4)

        stats_no_test_split = self.model.get_stats(ys_label, ys_label_pred,
                                                   split_test_indices_for_perf_ci=False)

        # check that the performance metric distributions are not in these stats dict
        assert 'SRCC_across_test_splits_distribution' not in stats_no_test_split, \
            'SRCC across_test_splits distribution should not exist.'
        assert 'PCC_across_test_splits_distribution' not in stats_no_test_split, \
            'PCC across_test_splits distribution should not exist.'
        assert 'RMSE_across_test_splits_distribution' not in stats_no_test_split, \
            'RMSE across_test_splits distribution should not exist.'

    def test_train_predict_randomforest(self):

        # random forest don't need proper data normalization

        xs = SklearnRandomForestTrainTestModel.get_xs_from_results(self.features, [0, 1, 2])
        ys = SklearnRandomForestTrainTestModel.get_ys_from_results(self.features, [0, 1, 2])
        xys = SklearnRandomForestTrainTestModel.get_xys_from_results(self.features, [0, 1, 2])

        model = SklearnRandomForestTrainTestModel({'norm_type': 'normalize',
                                                   'n_estimators': 10,
                                                   'random_state': 0}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.051804171170643766, places=4)

        model = SklearnRandomForestTrainTestModel({'norm_type': 'clip_0to1',
                                                   'n_estimators': 10,
                                                   'random_state': 0}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.051804171170643752, places=4)

        model = SklearnRandomForestTrainTestModel(
            {'norm_type': 'custom_clip_0to1',
             'n_estimators': 10,
             'custom_clip_0to1_map': {
                'Moment_noref_feature_1st_score': [0.0, 100.0],
              },
             'random_state': 0
             }, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.051804171170643752, places=4)

        model = SklearnRandomForestTrainTestModel({'norm_type': 'clip_minus1to1',
                                                   'n_estimators': 10,
                                                   'random_state': 0}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.051804171170643752, places=4)

        model = SklearnRandomForestTrainTestModel({'norm_type': 'none',
                                                   'n_estimators': 10,
                                                   'random_state': 0}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.051804171170643752, places=4)

    def test_train_predict_linearregression(self):

        # linear regression doesn't need proper data normalization

        xs = SklearnLinearRegressionTrainTestModel.get_xs_from_results(self.features, [0, 1, 2, 3, 4, 5])
        ys = SklearnLinearRegressionTrainTestModel.get_ys_from_results(self.features, [0, 1, 2, 3, 4, 5])
        xys = SklearnLinearRegressionTrainTestModel.get_xys_from_results(self.features, [0, 1, 2, 3, 4, 5])

        model = SklearnLinearRegressionTrainTestModel({'norm_type': 'normalize'}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)

        self.assertAlmostEqual(result['RMSE'], 0.49489849608079006, places=4)

    def test_train_predict_extratrees(self):

        # extra trees don't need proper data normalization

        xs = SklearnExtraTreesTrainTestModel.get_xs_from_results(self.features, [0, 1, 2])
        ys = SklearnExtraTreesTrainTestModel.get_ys_from_results(self.features, [0, 1, 2])
        xys = SklearnExtraTreesTrainTestModel.get_xys_from_results(self.features, [0, 1, 2])

        model = SklearnExtraTreesTrainTestModel({'norm_type': 'normalize',
                                                 'n_estimators': 10,
                                                 'random_state': 0}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.042867322777879642, places=4)

        model = SklearnExtraTreesTrainTestModel({'norm_type': 'clip_0to1',
                                                 'n_estimators': 10,
                                                 'random_state': 0}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.042867322777879642, places=4)

        model = SklearnExtraTreesTrainTestModel(
            {'norm_type': 'custom_clip_0to1',
             'n_estimators': 10,
             'custom_clip_0to1_map': {
                 'Moment_noref_feature_1st_score': [0.0, 100.0],
             },
             'random_state': 0
             }, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.042867322777879642, places=4)

        model = SklearnExtraTreesTrainTestModel({'norm_type': 'clip_minus1to1',
                                                 'n_estimators': 10,
                                                 'random_state': 0}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.042867322777879642, places=4)

        model = SklearnExtraTreesTrainTestModel({'norm_type': 'none',
                                                 'n_estimators': 10,
                                                 'random_state': 0,}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.042867322777879642, places=4)

    def test_train_logistic_fit_5PL(self):    
        xs = Logistic5PLRegressionTrainTestModel.get_xs_from_results(self.features, [0, 1, 2, 3, 4, 5], features=['Moment_noref_feature_1st_score'])
        ys = Logistic5PLRegressionTrainTestModel.get_ys_from_results(self.features, [0, 1, 2, 3, 4, 5])

        xys = {}
        xys.update(xs)
        xys.update(ys)
        
        model = Logistic5PLRegressionTrainTestModel({'norm_type': 'clip_0to1'}, None)
        model.train(xys)
        result = model.evaluate(xs, ys)

        self.assertAlmostEqual(result['RMSE'], 0.3603374311919728, places=4)


class TrainTestModelWithDisYRawVideoExtractorTest(unittest.TestCase):

    def setUp(self):

        train_dataset_path = VmafConfig.test_resource_path('test_image_dataset_diffdim2.py')
        train_dataset = import_python_file(train_dataset_path)
        train_assets = read_dataset(train_dataset)

        self.h5py_filepath = VmafConfig.workdir_path('test.hdf5')
        self.h5py_file = DisYUVRawVideoExtractor.open_h5py_file(self.h5py_filepath)
        optional_dict2 = {'h5py_file': self.h5py_file}

        runner = DisYUVRawVideoExtractor(
            train_assets,
            None,
            fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict=None,
            optional_dict2=optional_dict2,
        )
        runner.run(parallelize=False)  # CAN ONLY USE SERIAL MODE FOR DisYRawVideoExtractor
        self.features = runner.results

        self.model_filename = VmafConfig.workspace_path("model", "test_save_load.pkl")

    def tearDown(self):
        if hasattr(self, 'h5py_file'):
            DisYUVRawVideoExtractor.close_h5py_file(self.h5py_file)
        if os.path.exists(self.h5py_filepath):
            os.remove(self.h5py_filepath)
        if os.path.exists(self.model_filename):
            os.remove(self.model_filename)

    def test_extracted_features(self):
        self.assertAlmostEqual(np.mean(self.features[0]['dis_y']), 160.617204551784, places=4)
        self.assertAlmostEqual(np.mean(self.features[3]['dis_y']), 128.57907008374298, places=4)
        self.assertAlmostEqual(np.mean(self.features[0]['dis_u']), 135.38041204396345, places=4)
        self.assertAlmostEqual(np.mean(self.features[3]['dis_u']), 125.05257737968019, places=4)
        self.assertAlmostEqual(np.mean(self.features[0]['dis_v']), 125.26229752397977, places=4)
        self.assertAlmostEqual(np.mean(self.features[3]['dis_v']), 124.01385353721803, places=4)

    def test_get_xs_ys_consistent_wh(self):
        xs = MomentRandomForestTrainTestModel.get_xs_from_results(self.features, [0, 1, 2])
        ys = MomentRandomForestTrainTestModel.get_ys_from_results(self.features, [0, 1, 2])
        xys = MomentRandomForestTrainTestModel.get_xys_from_results(self.features, [0, 1, 2])

        expected_ys = {'label': [2.5, 3.9, 5.0], 'content_id': [0, 1, 2]}

        self.assertEqual(xs['dis_y'][0].shape, (1, 321, 481))
        self.assertEqual(xs['dis_y'][1].shape, (1, 321, 481))
        self.assertEqual(xs['dis_y'][2].shape, (1, 321, 481))
        self.assertEqual(xs['dis_u'][0].shape, (1, 321, 481))
        self.assertEqual(xs['dis_u'][1].shape, (1, 321, 481))
        self.assertEqual(xs['dis_u'][2].shape, (1, 321, 481))
        self.assertEqual(xs['dis_v'][0].shape, (1, 321, 481))
        self.assertEqual(xs['dis_v'][1].shape, (1, 321, 481))
        self.assertEqual(xs['dis_v'][2].shape, (1, 321, 481))
        self.assertAlmostEqual(np.mean(xs['dis_y']), 128.26146851380497, places=4)
        self.assertEqual(list(ys['label']), expected_ys['label'])
        self.assertEqual(list(ys['content_id']), expected_ys['content_id'])

        self.assertAlmostEqual(np.mean(xys['dis_y']), 128.26146851380497, places=4)
        self.assertEqual(list(xys['label']), expected_ys['label'])
        self.assertEqual(list(xys['content_id']), expected_ys['content_id'])

    def test_get_xs_ys_inconsistent_wh(self):
        xs = MomentRandomForestTrainTestModel.get_xs_from_results(self.features, [0, 8])
        ys = MomentRandomForestTrainTestModel.get_ys_from_results(self.features, [0, 8])
        xys = MomentRandomForestTrainTestModel.get_xys_from_results(self.features, [0, 8])

        expected_ys = {'label': [2.5, 4.7], 'content_id': [0, 13]}

        self.assertEqual(xs['dis_y'][0].shape, (1, 321, 481))
        self.assertEqual(xs['dis_y'][1].shape, (1, 486, 720))
        self.assertEqual(xs['dis_u'][0].shape, (1, 321, 481))
        self.assertEqual(xs['dis_u'][1].shape, (1, 486, 720))
        self.assertEqual(xs['dis_v'][0].shape, (1, 321, 481))
        self.assertEqual(xs['dis_v'][1].shape, (1, 486, 720))
        self.assertAlmostEqual(np.mean(xs['dis_y'][0]), 160.617204551784, places=4)
        self.assertAlmostEqual(np.mean(xs['dis_y'][1]), 35.8163894604481, places=4)
        self.assertEqual(list(ys['label']), expected_ys['label'])
        self.assertEqual(list(ys['content_id']), expected_ys['content_id'])

        self.assertEqual(xys['dis_y'][0].shape, (1, 321, 481))
        self.assertEqual(xys['dis_y'][1].shape, (1, 486, 720))
        self.assertEqual(xys['dis_u'][0].shape, (1, 321, 481))
        self.assertEqual(xys['dis_u'][1].shape, (1, 486, 720))
        self.assertEqual(xys['dis_v'][0].shape, (1, 321, 481))
        self.assertEqual(xys['dis_v'][1].shape, (1, 486, 720))
        self.assertEqual(list(xys['label']), expected_ys['label'])
        self.assertEqual(list(xys['content_id']), expected_ys['content_id'])

    def test_train_predict(self):
        xs = MomentRandomForestTrainTestModel.get_xs_from_results(self.features)
        ys = MomentRandomForestTrainTestModel.get_ys_from_results(self.features)
        xys = MomentRandomForestTrainTestModel.get_xys_from_results(self.features)

        # using dis_y only
        del xs['dis_u']
        del xs['dis_v']
        del xys['dis_u']
        del xys['dis_v']

        model = MomentRandomForestTrainTestModel({'norm_type': 'normalize',
                                                  'n_estimators': 10,
                                                  'random_state': 0})
        model.train(xys)

        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.17634739353518517, places=4)

    def test_train_save_load_predict(self):

        xs = MomentRandomForestTrainTestModel.get_xs_from_results(self.features)
        ys = MomentRandomForestTrainTestModel.get_ys_from_results(self.features)
        xys = MomentRandomForestTrainTestModel.get_xys_from_results(self.features)

        # using dis_y only
        del xs['dis_u']
        del xs['dis_v']
        del xys['dis_u']
        del xys['dis_v']

        model = MomentRandomForestTrainTestModel({'norm_type': 'normalize',
                                                  'n_estimators': 10,
                                                  'random_state': 0})
        model.train(xys)

        model.to_file(self.model_filename)
        self.assertTrue(os.path.exists(self.model_filename))

        loaded_model = TrainTestModel.from_file(self.model_filename)

        result = loaded_model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.17634739353518517, places=4)

    def test_train_predict_using_yuv(self):
        xs = MomentRandomForestTrainTestModel.get_xs_from_results(self.features)
        ys = MomentRandomForestTrainTestModel.get_ys_from_results(self.features)
        xys = MomentRandomForestTrainTestModel.get_xys_from_results(self.features)

        model = MomentRandomForestTrainTestModel({'norm_type': 'normalize',
                                                  'n_estimators': 10,
                                                  'random_state': 0})
        model.train(xys)

        result = model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.51128487038576109, places=4)


class TrainTestModelTestJson(unittest.TestCase):

    def setUp(self):

        train_dataset_path = VmafConfig.test_resource_path('test_image_dataset_diffdim2.py')
        train_dataset = import_python_file(train_dataset_path)
        train_assets = read_dataset(train_dataset)

        runner = MomentNorefFeatureExtractor(
            train_assets,
            None,
            fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict=None,
            optional_dict2=None,
        )
        runner.run(parallelize=False)
        self.features = runner.results

        self.model_filename_json = VmafConfig.workspace_path("model", "test_save_load.json")

    def tearDown(self):
        if hasattr(self, 'model'):
            self.model.delete(self.model_filename_json, format='json')

    def test_train_save_load_predict_libsvmnusvr_json(self):

        xs = LibsvmNusvrTrainTestModel.get_xs_from_results(self.features)
        ys = LibsvmNusvrTrainTestModel.get_ys_from_results(self.features)
        xys = LibsvmNusvrTrainTestModel.get_xys_from_results(self.features)

        self.model = LibsvmNusvrTrainTestModel({'norm_type': 'normalize'}, None)
        self.model.train(xys)

        self.model.to_file(self.model_filename_json, format='json')
        self.assertTrue(os.path.exists(self.model_filename_json))
        self.assertFalse(os.path.exists(self.model_filename_json + '.model'))

        loaded_model = LibsvmNusvrTrainTestModel.from_file(
            self.model_filename_json, logger=None, format='json')

        result = self.model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.62263086620058783, places=4)

        # loaded model generates slight numerical difference
        result = loaded_model.evaluate(xs, ys)
        self.assertAlmostEqual(result['RMSE'], 0.62263139871631323, places=4)


if __name__ == '__main__':
    unittest.main(verbosity=2)



__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

import sys
import unittest

from vmaf.core.train_test_model import SklearnRandomForestTrainTestModel, LibsvmNusvrTrainTestModel, \
    SklearnExtraTreesTrainTestModel
from vmaf.core.cross_validation import ModelCrossValidation
from vmaf.config import VmafConfig
from vmaf.core.noref_feature_extractor import MomentNorefFeatureExtractor
from vmaf.routine import read_dataset
from vmaf.tools.misc import import_python_file, unroll_dict_of_lists


class CrossValidationTest(unittest.TestCase):

    def setUp(self):

        train_dataset_path = VmafConfig.test_resource_path('test_image_dataset_diffdim2.py')
        train_dataset = import_python_file(train_dataset_path)
        train_assets = read_dataset(train_dataset)

        runner = MomentNorefFeatureExtractor(
            train_assets,
            None,
            fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict=None,
            optional_dict2=None,
        )
        runner.run(parallelize=True)
        self.features = runner.results

    def test_run_cross_validation(self):

        train_test_model_class = SklearnRandomForestTrainTestModel
        model_param = {'norm_type': 'normalize',
                       'n_estimators': 10,
                       'random_state': 0}

        indices_train = range(9)
        indices_test = range(9)

        output = ModelCrossValidation.run_cross_validation(
            train_test_model_class, model_param, self.features,
            indices_train, indices_test)
        self.assertAlmostEqual(output['stats']['SRCC'], 0.93333333333333324, places=4)
        self.assertAlmostEqual(output['stats']['PCC'], 0.97754442316039469, places=4)
        self.assertAlmostEqual(output['stats']['KENDALL'], 0.83333333333333337, places=4)
        self.assertAlmostEqual(output['stats']['RMSE'], 0.17634739353518517, places=4)
        self.assertEqual(output['model'].TYPE, "RANDOMFOREST")

    def test_run_kfold_cross_validation_randomforest(self):

        train_test_model_class = SklearnRandomForestTrainTestModel
        model_param = {'norm_type': 'normalize',
                       'n_estimators': 10,
                       'random_state': 0}

        output = ModelCrossValidation.run_kfold_cross_validation(
            train_test_model_class, model_param, self.features, 3)

        self.assertAlmostEqual(output['aggr_stats']['SRCC'], 0.28452131897694583, places=4)
        self.assertAlmostEqual(output['aggr_stats']['PCC'], 0.1689046198483892, places=4)
        self.assertAlmostEqual(output['aggr_stats']['KENDALL'], 0.084515425472851652, places=4)
        self.assertAlmostEqual(output['aggr_stats']['RMSE'], 1.344683833136588, places=4)

    def test_run_kfold_cross_validation_extratrees(self):

        train_test_model_class = SklearnExtraTreesTrainTestModel
        model_param = {'norm_type': 'normalize', 'random_state': 0, 'n_estimators': 10}

        output = ModelCrossValidation.run_kfold_cross_validation(
            train_test_model_class, model_param, self.features, 3)

        self.assertAlmostEqual(output['aggr_stats']['SRCC'], 0.17320508075688773, places=4)
        self.assertAlmostEqual(output['aggr_stats']['PCC'], 0.33023719320146966, places=4)
        self.assertAlmostEqual(output['aggr_stats']['KENDALL'], 0.14907119849998599, places=4)
        self.assertAlmostEqual(output['aggr_stats']['RMSE'], 1.3279056191361394, places=4)

    def test_run_kfold_cross_validation_libsvmnusvr(self):

        train_test_model_class = LibsvmNusvrTrainTestModel
        model_param = {'norm_type': 'normalize'}

        output = ModelCrossValidation.run_kfold_cross_validation(
            train_test_model_class, model_param, self.features, 3)

        self.assertAlmostEqual(output['aggr_stats']['SRCC'], 0.31666666666666665, places=4)
        self.assertAlmostEqual(output['aggr_stats']['PCC'], 0.33103132578536021, places=4)
        self.assertAlmostEqual(output['aggr_stats']['KENDALL'], 0.27777777777777779, places=4)
        self.assertAlmostEqual(output['aggr_stats']['RMSE'], 1.2855099934718619, places=4)

    def test_run_kfold_cross_validation_with_list_input(self):

        train_test_model_class = SklearnRandomForestTrainTestModel
        model_param = {'norm_type': 'normalize',
                       'n_estimators': 10,
                       'random_state': 0}

        output = ModelCrossValidation.run_kfold_cross_validation(
            train_test_model_class, model_param, self.features,
            [[0, 3, 8], [2, 1, 5], [4, 6, 7]])

        self.assertAlmostEqual(output['aggr_stats']['SRCC'], 0.18333333333333335, places=4)
        self.assertAlmostEqual(output['aggr_stats']['PCC'], 0.35513638509959689, places=4)
        self.assertAlmostEqual(output['aggr_stats']['KENDALL'], 0.1111111111111111, places=3)
        self.assertAlmostEqual(output['aggr_stats']['RMSE'], 1.2740400878438387, places=3)

    def test_unroll_dict_of_lists(self):
        model_param_search_range = {'norm_type': ['normalize', 'clip_0to1'],
                                    'n_estimators': [10, 50], 'random_state': [0]}

        dicts = unroll_dict_of_lists(model_param_search_range)

        expected_dicts = [
         {'norm_type': 'normalize', 'n_estimators': 10, 'random_state': 0},
         {'norm_type': 'clip_0to1', 'n_estimators': 10, 'random_state': 0},
         {'norm_type': 'normalize', 'n_estimators': 50, 'random_state': 0},
         {'norm_type': 'clip_0to1', 'n_estimators': 50, 'random_state': 0},
        ]

        self.assertEqual(dicts, expected_dicts)

    @unittest.skipIf(sys.version_info < (3,), reason="For py3 only: py2 uses a different random seed.")
    def test_sample_model_param_list(self):
        import random
        random.seed(0)

        model_param_search_range = {'norm_type': ['normalize', 'clip_0to1'],
                                    'n_estimators': [10, 50], 'random_state': [0]}
        dicts = ModelCrossValidation._sample_model_param_list(model_param_search_range, 4)
        expected_dicts = [
            {'n_estimators': 50, 'norm_type': 'clip_0to1', 'random_state': 0},
            {'n_estimators': 50, 'norm_type': 'clip_0to1', 'random_state': 0},
            {'n_estimators': 50, 'norm_type': 'clip_0to1', 'random_state': 0},
            {'n_estimators': 10, 'norm_type': 'normalize', 'random_state': 0},
        ]
        for actual_dict, expected_dict in zip(dicts, expected_dicts):
            self.assertDictEqual(actual_dict, expected_dict)

        model_param_search_range = {'norm_type': ['normalize', 'clip_0to1'],
                                    'n_estimators': {'low': 10, 'high': 50, 'decimal': 0},
                                    'random_state': [0]}
        dicts = ModelCrossValidation._sample_model_param_list(
            model_param_search_range, 4)
        expected_dicts = [
            {'n_estimators': 15.0, 'norm_type': 'normalize', 'random_state': 0},
            {'n_estimators': 49.0, 'norm_type': 'normalize', 'random_state': 0},
            {'n_estimators': 13.0, 'norm_type': 'normalize', 'random_state': 0},
            {'n_estimators': 28.0, 'norm_type': 'normalize', 'random_state': 0},
        ]
        self.assertEqual(dicts, expected_dicts)

    def test_find_most_frequent_dict(self):
        dicts = [
         {'norm_type': 'normalize', 'n_estimators': 10, 'random_state': 0},
         {'norm_type': 'normalize', 'n_estimators': 50, 'random_state': 0},
         {'norm_type': 'clip_0to1', 'n_estimators': 10, 'random_state': 0},
         {'norm_type': 'clip_0to1', 'n_estimators': 50, 'random_state': 0},
         {'norm_type': 'clip_0to1', 'n_estimators': 50, 'random_state': 0},
        ]

        actual_dict, count = ModelCrossValidation._find_most_frequent_dict(dicts)
        expected_dict = {'norm_type': 'clip_0to1', 'n_estimators': 50, 'random_state': 0}
        expected_count = 2

        self.assertEqual(actual_dict, expected_dict)
        self.assertEqual(count, expected_count)

    def test_run_nested_kfold_cross_validation_randomforest(self):

        train_test_model_class = SklearnRandomForestTrainTestModel
        model_param_search_range = \
            {'norm_type': ['normalize'],
             'n_estimators': [10, 90],
             'max_depth': [None, 3],
             'random_state': [0]}

        output = ModelCrossValidation.run_nested_kfold_cross_validation(
            train_test_model_class, model_param_search_range, self.features, 3)

        self.assertAlmostEqual(output['aggr_stats']['SRCC'], 0.40167715620274708, places=4)
        self.assertAlmostEqual(output['aggr_stats']['PCC'], 0.11009919053282299, places=4)
        self.assertAlmostEqual(output['aggr_stats']['KENDALL'], 0.14085904245475275, places=4)
        self.assertAlmostEqual(output['aggr_stats']['RMSE'], 1.3681348274719265, places=4)

        expected_top_model_param = {
            'norm_type': 'normalize',
            'n_estimators': 10,
            'max_depth': None,
            'random_state': 0
        }
        expected_top_ratio = 0.6666666666666666
        self.assertEqual(output['top_model_param'], expected_top_model_param)
        self.assertEqual(output['top_ratio'], expected_top_ratio)

    def test_run_nested_kfold_cross_validation_libsvmnusvr(self):

        train_test_model_class = LibsvmNusvrTrainTestModel
        model_param_search_range = \
            {'norm_type': ['normalize', 'clip_0to1', 'clip_minus1to1'],
             'kernel': ['rbf'],
             'nu': [0.5],
             'C': [1, 2],
             'gamma': [0.0]
             }

        output = ModelCrossValidation.run_nested_kfold_cross_validation(
            train_test_model_class, model_param_search_range, self.features, 3)

        self.assertAlmostEqual(output['aggr_stats']['SRCC'], 0.30962614123961751, places=4)
        self.assertAlmostEqual(output['aggr_stats']['PCC'], -0.1535643705229309, places=4)
        self.assertAlmostEqual(output['aggr_stats']['KENDALL'], 0.14085904245475275, places=4)
        self.assertAlmostEqual(output['aggr_stats']['RMSE'], 1.5853397658781734, places=4)

        expected_top_model_param = {
            'norm_type': 'clip_0to1',
            'kernel': 'rbf',
            'nu': 0.5,
            'C': 1,
            'gamma': 0.0,
        }
        expected_top_ratio = 1.0

        self.assertEqual(output['top_model_param'], expected_top_model_param)
        self.assertEqual(output['top_ratio'], expected_top_ratio)

    def test_run_nested_kfold_cross_validation_with_list_input(self):

        train_test_model_class = SklearnRandomForestTrainTestModel
        model_param_search_range = \
            {'norm_type': ['none'],
             'n_estimators': [10, 90],
             'max_depth': [None, 3],
             'random_state': [0]
             }

        output = ModelCrossValidation.run_nested_kfold_cross_validation(
            train_test_model_class, model_param_search_range, self.features,
            [[0, 3, 2], [8, 6, 5], [4, 1, 7]])

        self.assertAlmostEqual(output['aggr_stats']['SRCC'], 0.26666666666666666, places=4)
        self.assertAlmostEqual(output['aggr_stats']['PCC'], 0.15272340058922063, places=4)
        self.assertAlmostEqual(output['aggr_stats']['KENDALL'], 0.22222222222222221, places=4)
        self.assertAlmostEqual(output['aggr_stats']['RMSE'], 1.452887116343635, places=4)

        expected_top_model_param = {'norm_type': 'none',
                                    'n_estimators': 10,
                                    'max_depth': None,
                                    'random_state': 0
                                    }
        expected_top_ratio = 0.6666666666666666
        self.assertEqual(output['top_model_param'], expected_top_model_param)
        self.assertEqual(output['top_ratio'], expected_top_ratio)


if __name__ == '__main__':
    unittest.main(verbosity=2)



dataset_name = 'example'

quality_width = 1920
quality_height = 1080

from vmaf.config import VmafConfig

ref_videos = [
    {'content_id': 0, 'path': VmafConfig.resource_path('yuv', 'checkerboard_1920_1080_10_3_0_0.yuv'), 'yuv_fmt': 'yuv420p'},
]

dis_videos = [
    {'content_id': 0, 'asset_id': 0, 'dmos': 100, 'path': VmafConfig.resource_path('yuv', 'checkerboard_1920_1080_10_3_0_0.yuv'), 'fps': 30, 'rebuf_indices': [0, 4, 15]}, # ref, make up fps and rebuf indices
]